Package: tabpfn
Title: Prior-Data Fitted Network Foundational Model for Tabular Data
Version: 0.2.0
Authors@R: c(
    person("Max", "Kuhn", , "max@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2402-136X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Provides a consistent API for classification and regression models 
    based on the 'TabPFN' model of Hollmann et al. (2025), "Accurate predictions 
    on small data with a tabular foundation model," Nature, 637(8045) 
    <doi:10.1038/s41586-024-08328-6>. The calculations are served via 'Python' 
    to train and predict the model. 
License: Apache License (>= 2)
URL: https://tabpfn.tidymodels.org,
        https://github.com/tidymodels/tabpfn
BugReports: https://github.com/tidymodels/tabpfn/issues
Depends: R (>= 4.1.0)
Imports: cli, dplyr, generics, hardhat, purrr, reticulate (>=
        1.41.0.1), rlang (>= 1.1.0), tibble
Suggests: covr, modeldata, recipes, spelling, testthat (>= 3.0.0),
        withr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-14 12:41:48 UTC; max
Author: Max Kuhn [aut, cre] (ORCID: <https://orcid.org/0000-0003-2402-136X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Max Kuhn <max@posit.co>
Repository: CRAN
Date/Publication: 2026-05-14 13:00:02 UTC
Built: R 4.5.2; ; 2026-05-14 15:01:12 UTC; unix

Package: tabpfn
Title: Prior-Data Fitted Network Foundational Model for Tabular Data
Version: 0.2.0
Authors@R: c(
    person("Max", "Kuhn", , "max@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2402-136X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Provides a consistent API for classification and regression models 
    based on the 'TabPFN' model of Hollmann et al. (2025), "Accurate predictions 
    on small data with a tabular foundation model," Nature, 637(8045) 
    <doi:10.1038/s41586-024-08328-6>. The calculations are served via 'Python' 
    to train and predict the model. 
License: Apache License (>= 2)
URL: https://tabpfn.tidymodels.org,
        https://github.com/tidymodels/tabpfn
BugReports: https://github.com/tidymodels/tabpfn/issues
Depends: R (>= 4.1.0)
Imports: cli, dplyr, generics, hardhat, purrr, reticulate (>=
        1.41.0.1), rlang (>= 1.1.0), tibble
Suggests: covr, modeldata, recipes, spelling, testthat (>= 3.0.0),
        withr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-14 12:41:48 UTC; max
Author: Max Kuhn [aut, cre] (ORCID: <https://orcid.org/0000-0003-2402-136X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Max Kuhn <max@posit.co>
Repository: CRAN
Date/Publication: 2026-05-14 13:00:02 UTC
Built: R 4.5.2; ; 2026-05-14 14:22:23 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(augment,tab_pfn)
S3method(predict,tab_pfn)
S3method(predict,tabpfn.classifier.TabPFNClassifier)
S3method(predict,tabpfn.regressor.TabPFNRegressor)
S3method(print,control_tab_pfn)
S3method(print,tab_pfn)
S3method(tab_pfn,data.frame)
S3method(tab_pfn,default)
S3method(tab_pfn,formula)
S3method(tab_pfn,matrix)
S3method(tab_pfn,recipe)
export(augment)
export(control_tab_pfn)
export(is_tab_pfn_installed)
export(tab_pfn)
export(tabpfn_download_models)
export(tabpfn_list_versions)
import(rlang)
importFrom(generics,augment)
importFrom(stats,predict)

# tabpfn

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/tabpfn)](https://CRAN.R-project.org/package=tabpfn)
[![R-CMD-check](https://github.com/tidymodels/tabpfn/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidymodels/tabpfn/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/tidymodels/tabpfn/branch/main/graph/badge.svg)](https://app.codecov.io/gh/tidymodels/tabpfn?branch=main)
<!-- badges: end -->

tabpfn, meaning prior fitted networks for tabular data, is a
deep-learning model. See:

- [*Transformers Can Do Bayesian
  Inference*](https://arxiv.org/abs/2112.10510) (arXiv, 2021)
- [*TabPFN: A Transformer That Solves Small Tabular Classification
  Problems in a Second*](https://arxiv.org/abs/2207.01848) (arXiv, 2022)
- [*Accurate predictions on small data with a tabular foundation
  model*](https://scholar.google.com/scholar?hl=en&as_sdt=0%2C7&q=%22Accurate+predictions+on+small+data+with+a+tabular+foundation+model%22)
  (Nature, 2025)

This R package is a wrapper of the [Python
library](https://github.com/PriorLabs/tabpfn) via reticulate. It has an
idiomatic R syntax using standard S3 methods.

## Installation

You can download the package from CRAN via:

``` r
install.packages("tabpfn")
```

or you can install the development version of tabpfn like so:

``` r
require(pak)
pak(c("tidymodels/tabpfn"), ask = FALSE)
```

You’ll need a Python virtual environment to access the underlying
library. After installing the R package, tabpfn will install the
required Python bits when you first fit a model:

    > library(tabpfn)
    >
    > predictors <- mtcars[, -1]
    > outcome <- mtcars[, 1]
    >
    > # XY interface
    > mod <- tab_pfn(predictors, outcome)
    Downloading uv...Done!
    Downloading cpython-3.12.12 (download) (15.9MiB)
     Downloading cpython-3.12.12 (download)
    Downloading setuptools (1.1MiB)
    Downloading scikit-learn (8.2MiB)
    Downloading numpy (4.9MiB)

    <downloading and installing more packages>

     Downloading llvmlite
     Downloading torch
    Installed 58 packages in 350ms
    > mod
    tabpfn Regression Model

    Training set
    i 32 data points
    i 10 predictors

## Example

After loading the package:

``` r
library(tabpfn)
```

we can fit a model via the standard x/y interface.

``` r
set.seed(364)
reg_mod <- tab_pfn(mtcars[1:25, -1], mtcars$mpg[1:25])
reg_mod
#> TabPFN Regression Model
#> Training set
#> ℹ 25 data points
#> ℹ 10 predictors
```

There are also formula and recipes interfaces.

Prediction follows the usual S3 `predict()` method:

``` r
predict(reg_mod, mtcars[26:32, -1])
#> # A tibble: 7 × 1
#>   .pred
#>   <dbl>
#> 1  31.4
#> 2  24.3
#> 3  24.8
#> 4  16.4
#> 5  18.9
#> 6  14.4
#> 7  22.5
```

tabpfn follows the tidymodels prediction convention: a data frame is
always returned with a standard set of column names.

For a classification model, the outcome should always be a factor
vector. For example, using these data from the modeldata package:

``` r
library(modeldata)
#> 
#> Attaching package: 'modeldata'
#> The following object is masked from 'package:datasets':
#> 
#>     penguins
library(ggplot2)

two_cls_train <- parabolic[1:400,  ]
two_cls_val   <- parabolic[401:500,]
grid <- expand.grid(X1 = seq(-5.1, 5.0, length.out = 25), 
                    X2 = seq(-5.5, 4.0, length.out = 25))

set.seed(3824)
cls_mod <- tab_pfn(class ~ ., data = two_cls_train)

grid_pred <- predict(cls_mod, grid)
grid_pred
#> # A tibble: 625 × 3
#>    .pred_Class1 .pred_Class2 .pred_class
#>           <dbl>        <dbl> <fct>      
#>  1        0.997      0.00273 Class1     
#>  2        0.998      0.00217 Class1     
#>  3        0.998      0.00182 Class1     
#>  4        0.998      0.00155 Class1     
#>  5        0.998      0.00167 Class1     
#>  6        0.998      0.00222 Class1     
#>  7        0.996      0.00438 Class1     
#>  8        0.989      0.0109  Class1     
#>  9        0.948      0.0522  Class1     
#> 10        0.745      0.255   Class1     
#> # ℹ 615 more rows
```

The fit looks fairly good when shown with out-of-sample data:

``` r
cbind(grid, grid_pred) |>
  ggplot(aes(X1, X2)) +
  geom_point(
    data = two_cls_val,
    aes(col = class, pch = class),
    alpha = 3 / 4,
    cex = 3
  ) +
  geom_contour(
    aes(z = .pred_Class1),
    breaks = 1 / 2,
    col = "black",
    linewidth = 1
  ) +
  coord_equal(ratio = 1)
```

<img src="man/figures/README-boundaries-1.png" alt="" width="70%" style="display: block; margin: auto;" />

## License

[PriorLabs](https://priorlabs.ai/) created the model. Starting with
version 2.5, using TabPFN requires accepting the model license and
setting a token. Each model version (v2.5, v2.6, etc.) has its own
license that must be accepted individually.

To get access, visit <https://ux.priorlabs.ai>, go to the **Licenses**
tab (1), and accept the license for each model version you intend to use
(2). Then set the `TABPFN_TOKEN` environment variable with the token
from your account. Users who already have `TABPFN_TOKEN` set can use
TabPFN v2 without any additional steps.

<figure>
<img src="man/figures/license.png"
alt="Screenshot of the PriorLabs UX portal Licenses page" />
<figcaption aria-hidden="true">Screenshot of the PriorLabs UX portal
Licenses page</figcaption>
</figure>

Also, the model is most effective when a GPU is available (by an order
of magnitude or two). This may seem obvious to anyone already working
with deep learning models, but it is a fairly new requirement for those
strictly working with traditional tabular data models.

## Code of Conduct

Please note that the tabpfn project is released with a [Contributor Code
of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

# TabPFN

[![PyPI version](https://badge.fury.io/py/tabpfn.svg)](https://badge.fury.io/py/tabpfn)
[![Downloads](https://pepy.tech/badge/tabpfn)](https://pepy.tech/project/tabpfn)
[![Discord](https://img.shields.io/discord/1285598202732482621?color=7289da&label=Discord&logo=discord&logoColor=ffffff)](https://discord.gg/BHnX2Ptf4j)
[![Documentation](https://img.shields.io/badge/docs-priorlabs.ai-blue)](https://priorlabs.ai/docs)
[![colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/PriorLabs/TabPFN/blob/main/examples/notebooks/TabPFN_Demo_Local.ipynb)
[![Python Versions](https://img.shields.io/badge/python-3.9%20%7C%203.10%20%7C%203.11%20%7C%203.12%20%7C%203.13-blue)](https://pypi.org/project/tabpfn/)

<img src="https://github.com/PriorLabs/tabpfn-extensions/blob/main/tabpfn_summary.webp" width="80%" alt="TabPFN Summary">

## Quick Start

### Interactive Notebook Tutorial
> [!TIP]
>
> Dive right in with our interactive Colab notebook! It's the best way to get a hands-on feel for TabPFN, walking you through installation, classification, and regression examples.
>
> [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/PriorLabs/TabPFN/blob/main/examples/notebooks/TabPFN_Demo_Local.ipynb)

> ⚡ **GPU Recommended**:
> For optimal performance, use a GPU (even older ones with ~8GB VRAM work well; 16GB needed for some large datasets).
> On CPU, only small datasets (≲1000 samples) are feasible.
> No GPU? Use our free hosted inference via [TabPFN Client](https://github.com/PriorLabs/tabpfn-client).

### Installation
Official installation (pip)
```bash
pip install tabpfn
```
OR installation from source
```bash
pip install "tabpfn @ git+https://github.com/PriorLabs/TabPFN.git"
```
OR local development installation
```bash

git clone https://github.com/PriorLabs/TabPFN.git --depth 1
pip install -e "TabPFN[dev]"
```

### Basic Usage

#### Classification
```python
from sklearn.datasets import load_breast_cancer
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.model_selection import train_test_split

from tabpfn import TabPFNClassifier
from tabpfn.constants import ModelVersion

# Load data
X, y = load_breast_cancer(return_X_y=True)
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.5, random_state=42)

# Initialize a classifier
clf = TabPFNClassifier()  # Uses TabPFN 2.5 weights, finetuned on real data.
# To use TabPFN v2:
# clf = TabPFNClassifier.create_default_for_version(ModelVersion.V2)
clf.fit(X_train, y_train)


# Predict probabilities
prediction_probabilities = clf.predict_proba(X_test)
print("ROC AUC:", roc_auc_score(y_test, prediction_probabilities[:, 1]))

# Predict labels
predictions = clf.predict(X_test)
print("Accuracy", accuracy_score(y_test, predictions))
```

#### Regression
```python
from sklearn.datasets import fetch_openml
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.model_selection import train_test_split

from tabpfn import TabPFNRegressor
from tabpfn.constants import ModelVersion

# Load Boston Housing data
df = fetch_openml(data_id=531, as_frame=True)  # Boston Housing dataset
X = df.data
y = df.target.astype(float)  # Ensure target is float for regression

# Train-test split
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.5, random_state=42)

# Initialize the regressor
regressor = TabPFNRegressor()  # Uses TabPFN-2.5 weights, trained on synthetic data only.
# To use TabPFN v2:
# regressor = TabPFNRegressor.create_default_for_version(ModelVersion.V2)
regressor.fit(X_train, y_train)

# Predict on the test set
predictions = regressor.predict(X_test)

# Evaluate the model
mse = mean_squared_error(y_test, predictions)
r2 = r2_score(y_test, predictions)

print("Mean Squared Error (MSE):", mse)
print("R² Score:", r2)
```

## TabPFN Ecosystem

Choose the right TabPFN implementation for your needs:

- **[TabPFN Client](https://github.com/priorlabs/tabpfn-client)**
  Simple API client for using TabPFN via cloud-based inference.

- **[TabPFN Extensions](https://github.com/priorlabs/tabpfn-extensions)**
  A powerful companion repository packed with advanced utilities, integrations, and features - great place to contribute:

  -  **`interpretability`**: Gain insights with SHAP-based explanations, feature importance, and selection tools.
  -  **`unsupervised`**: Tools for outlier detection and synthetic tabular data generation.
  -  **`embeddings`**: Extract and use TabPFN’s internal learned embeddings for downstream tasks or analysis.
  -  **`many_class`**: Handle multi-class classification problems that exceed TabPFN's built-in class limit.
  -  **`rf_pfn`**: Combine TabPFN with traditional models like Random Forests for hybrid approaches.
  -  **`hpo`**: Automated hyperparameter optimization tailored to TabPFN.
  -  **`post_hoc_ensembles`**: Boost performance by ensembling multiple TabPFN models post-training.

  To install:
  ```bash
  git clone https://github.com/priorlabs/tabpfn-extensions.git
  pip install -e tabpfn-extensions
  ```

- **[TabPFN (this repo)](https://github.com/priorlabs/tabpfn)**
  Core implementation for fast and local inference with PyTorch and CUDA support.

- **[TabPFN UX](https://ux.priorlabs.ai)**
  No-code graphical interface to explore TabPFN capabilities—ideal for business users and prototyping.

## TabPFN Workflow at a Glance
Follow this decision tree to build your model and choose the right extensions from our ecosystem. It walks you through critical questions about your data, hardware, and performance needs, guiding you to the best solution for your specific use case.

```mermaid
---
config:
  theme: 'default'
  themeVariables:
    edgeLabelBackground: 'white'
---
graph LR
    %% 1. DEFINE COLOR SCHEME & STYLES
    classDef default fill:#fff,stroke:#333,stroke-width:2px,color:#333;
    classDef start_node fill:#e8f5e9,stroke:#43a047,stroke-width:2px,color:#333;
    classDef process_node fill:#e0f2f1,stroke:#00796b,stroke-width:2px,color:#333;
    classDef decision_node fill:#fff8e1,stroke:#ffa000,stroke-width:2px,color:#333;

    style Infrastructure fill:#fff,stroke:#ccc,stroke-width:5px;
    style Unsupervised fill:#fff,stroke:#ccc,stroke-width:5px;
    style Data fill:#fff,stroke:#ccc,stroke-width:5px;
    style Performance fill:#fff,stroke:#ccc,stroke-width:5px;
    style Interpretability fill:#fff,stroke:#ccc,stroke-width:5px;

    %% 2. DEFINE GRAPH STRUCTURE
    subgraph Infrastructure
        start((Start)) --> gpu_check["GPU available?"];
        gpu_check -- Yes --> local_version["Use TabPFN<br/>(local PyTorch)"];
        gpu_check -- No --> api_client["Use TabPFN-Client<br/>(cloud API)"];
        task_type["What is<br/>your task?"]
    end

    local_version --> task_type
    api_client --> task_type

    end_node((Workflow<br/>Complete));

    subgraph Unsupervised
        unsupervised_type["Select<br/>Unsupervised Task"];
        unsupervised_type --> imputation["Imputation"]
        unsupervised_type --> data_gen["Data<br/>Generation"];
        unsupervised_type --> tabebm["Data<br/>Augmentation"];
        unsupervised_type --> density["Outlier<br/>Detection"];
        unsupervised_type --> embedding["Get<br/>Embeddings"];
    end


    subgraph Data
        data_check["Data Checks"];
        model_choice["Samples > 10k or<br/>Classes > 10?"]
        data_check -- "Table Contains Text Data?" --> api_backend_note["Note: API client has<br/>native text support"];
        api_backend_note --> model_choice;
        data_check -- "Time-Series Data?" --> ts_features["Use Time-Series<br/>Features"];
        ts_features --> model_choice;
        data_check -- "Purely Tabular" --> model_choice;
        model_choice -- "No" --> finetune_check;
        model_choice -- "Yes, >10k samples" --> subsample["Large Datasets Guide<br/>"];
        model_choice -- "Yes, >10 classes" --> many_class["Many-Class<br/>Method"];
    end

    subgraph Performance
        finetune_check["Need<br/>Finetuning?"];
        performance_check["Need Even Better Performance?"];
        speed_check["Need faster inference<br/>at prediction time?"];
        kv_cache["Enable KV Cache<br/>(fit_mode='fit_with_cache')<br/><small>Faster predict; +Memory ~O(N×F)</small>"];
        tuning_complete["Tuning Complete"];

        finetune_check -- Yes --> finetuning["Finetuning"];
        finetune_check -- No --> performance_check;

        finetuning --> performance_check;

        performance_check -- No --> tuning_complete;
        performance_check -- Yes --> hpo["HPO"];
        performance_check -- Yes --> post_hoc["Post-Hoc<br/>Ensembling"];
        performance_check -- Yes --> more_estimators["More<br/>Estimators"];
        performance_check -- Yes --> speed_check;

        speed_check -- Yes --> kv_cache;
        speed_check -- No --> tuning_complete;

        hpo --> tuning_complete;
        post_hoc --> tuning_complete;
        more_estimators --> tuning_complete;
        kv_cache --> tuning_complete;
    end

    subgraph Interpretability

        tuning_complete --> interpretability_check;

        interpretability_check["Need<br/>Interpretability?"];

        interpretability_check --> feature_selection["Feature Selection"];
        interpretability_check --> partial_dependence["Partial Dependence Plots"];
        interpretability_check --> shapley["Explain with<br/>SHAP"];
        interpretability_check --> shap_iq["Explain with<br/>SHAP IQ"];
        interpretability_check -- No --> end_node;

        feature_selection --> end_node;
        partial_dependence --> end_node;
        shapley --> end_node;
        shap_iq --> end_node;

    end

    %% 3. LINK SUBGRAPHS AND PATHS
    task_type -- "Classification or Regression" --> data_check;
    task_type -- "Unsupervised" --> unsupervised_type;

    subsample --> finetune_check;
    many_class --> finetune_check;

    %% 4. APPLY STYLES
    class start,end_node start_node;
    class local_version,api_client,imputation,data_gen,tabebm,density,embedding,api_backend_note,ts_features,subsample,many_class,finetuning,feature_selection,partial_dependence,shapley,shap_iq,hpo,post_hoc,more_estimators,kv_cache process_node;
    class gpu_check,task_type,unsupervised_type,data_check,model_choice,finetune_check,interpretability_check,performance_check,speed_check decision_node;
    class tuning_complete process_node;

    %% 5. ADD CLICKABLE LINKS (INCLUDING KV CACHE EXAMPLE)
    click local_version "https://github.com/PriorLabs/TabPFN" "TabPFN Backend Options" _blank
    click api_client "https://github.com/PriorLabs/tabpfn-client" "TabPFN API Client" _blank
    click api_backend_note "https://github.com/PriorLabs/tabpfn-client" "TabPFN API Backend" _blank
    click unsupervised_type "https://github.com/PriorLabs/tabpfn-extensions" "TabPFN Extensions" _blank
    click imputation "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/unsupervised/imputation.py" "TabPFN Imputation Example" _blank
    click data_gen "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/unsupervised/generate_data.py" "TabPFN Data Generation Example" _blank
    click tabebm "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/tabebm/tabebm_augment_real_world_data.ipynb" "TabEBM Data Augmentation Example" _blank
    click density "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/unsupervised/density_estimation_outlier_detection.py" "TabPFN Density Estimation/Outlier Detection Example" _blank
    click embedding "https://github.com/PriorLabs/tabpfn-extensions/tree/main/examples/embedding" "TabPFN Embedding Example" _blank
    click ts_features "https://github.com/PriorLabs/tabpfn-time-series" "TabPFN Time-Series Example" _blank
    click many_class "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/many_class/many_class_classifier_example.py" "Many Class Example" _blank
    click finetuning "https://github.com/PriorLabs/TabPFN/blob/main/examples/finetune_classifier.py" "Finetuning Example" _blank
    click feature_selection "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/interpretability/feature_selection.py" "Feature Selection Example" _blank
    click partial_dependence "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/interpretability/pdp_example.py" "Partial Dependence Plots Example" _blank
    click shapley "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/interpretability/shap_example.py" "Shapley Values Example" _blank
    click shap_iq "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/interpretability/shapiq_example.py" "SHAP IQ Example" _blank
    click post_hoc "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/phe/phe_example.py" "Post-Hoc Ensemble Example" _blank
    click hpo "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/hpo/tuned_tabpfn.py" "HPO Example" _blank
    click subsample "https://github.com/PriorLabs/tabpfn-extensions/blob/main/examples/large_datasets/large_datasets_example.py" "Large Datasets Example" _blank
    click kv_cache "https://github.com/PriorLabs/TabPFN/blob/main/examples/kv_cache_fast_prediction.py" "KV Cache Fast Prediction Example" _blank

```

## License

The TabPFN-2.5 model weights are licensed under a [non-commercial license](https://huggingface.co/Prior-Labs/tabpfn_2_5/blob/main/LICENSE). These are used by default.

The code and TabPFN-2 model weights are licensed under Prior Labs License (Apache 2.0 with additional attribution requirement): [here](https://priorlabs.ai/tabpfn-license/). To use the v2 model weights, instantiate your model as follows:

```
from tabpfn.constants import ModelVersion
tabpfn_v2 = TabPFNRegressor.create_default_for_version(ModelVersion.V2)
```



## Join Our Community

We're building the future of tabular machine learning and would love your involvement:

1. **Connect & Learn**:
   - Join our [Discord Community](https://discord.gg/VJRuU3bSxt)
   - Read our [Documentation](https://priorlabs.ai/docs)
   - Check out [GitHub Issues](https://github.com/priorlabs/tabpfn/issues)

2. **Contribute**:
   - Report bugs or request features
   - Submit pull requests (please make sure to open an issue discussing the feature/bug first if none exists)
   - Share your research and use cases

3. **Stay Updated**: Star the repo and join Discord for the latest updates

## Citation

You can read our paper explaining TabPFN [here](https://doi.org/10.1038/s41586-024-08328-6).

```bibtex
@article{hollmann2025tabpfn,
 title={Accurate predictions on small data with a tabular foundation model},
 author={Hollmann, Noah and M{\"u}ller, Samuel and Purucker, Lennart and
         Krishnakumar, Arjun and K{\"o}rfer, Max and Hoo, Shi Bin and
         Schirrmeister, Robin Tibor and Hutter, Frank},
 journal={Nature},
 year={2025},
 month={01},
 day={09},
 doi={10.1038/s41586-024-08328-6},
 publisher={Springer Nature},
 url={https://www.nature.com/articles/s41586-024-08328-6},
}

@inproceedings{hollmann2023tabpfn,
  title={TabPFN: A transformer that solves small tabular classification problems in a second},
  author={Hollmann, Noah and M{\"u}ller, Samuel and Eggensperger, Katharina and Hutter, Frank},
  booktitle={International Conference on Learning Representations 2023},
  year={2023}
}
```



## ❓ FAQ

### **Usage & Compatibility**

**Q: What dataset sizes work best with TabPFN?**
A: TabPFN-2.5 is optimized for **datasets up to 50,000 rows**. For larger datasets, consider using **Random Forest preprocessing** or other extensions. See our [Colab notebook](https://colab.research.google.com/drive/154SoIzNW1LHBWyrxNwmBqtFAr1uZRZ6a#scrollTo=OwaXfEIWlhC8) for strategies.

**Q: Why can't I use TabPFN with Python 3.8?**
A: TabPFN requires **Python 3.9+** due to newer language features. Compatible versions: **3.9, 3.10, 3.11, 3.12, 3.13**.

### **Installation & Setup**

**Q: How do I get access to TabPFN-2.5?**

Visit [https://huggingface.co/Prior-Labs/tabpfn_2_5](https://huggingface.co/Prior-Labs/tabpfn_2_5) and accept the license terms. If access via huggingface is not an option for you, please contact us at [`sales@priorlabs.ai`](mailto:sales@priorlabs.ai).

Downloading the model requires your machine to be logged into Hugging Face. To do so, run `hf auth login` in your terminal, see the [huggingface documentation](https://huggingface.co/docs/huggingface_hub/en/quick-start#authentication) for details..

**Q: How do I use TabPFN without an internet connection?**

TabPFN automatically downloads model weights when first used. For offline usage:

**Using the Provided Download Script**

If you have the TabPFN repository, you can use the included script to download all models (including ensemble variants):

```bash
# After installing TabPFN
python scripts/download_all_models.py
```

This script will download the main classifier and regressor models, as well as all ensemble variant models to your system's default cache directory.

**Manual Download**

1. Download the model files manually from HuggingFace:
   - Classifier: [tabpfn-v2.5-classifier-v2.5_default.ckpt](https://huggingface.co/Prior-Labs/tabpfn_2_5/blob/main/tabpfn-v2.5-classifier-v2.5_default.ckpt) (Note: the classifier default uses the model fine-tuned on real data).
   - Regressor: [tabpfn-v2.5-regressor-v2.5_default.ckpt](https://huggingface.co/Prior-Labs/tabpfn_2_5/blob/main/tabpfn-v2.5-regressor-v2.5_default.ckpt)

2. Place the file in one of these locations:
   - Specify directly: `TabPFNClassifier(model_path="/path/to/model.ckpt")`
   - Set environment variable: `export TABPFN_MODEL_CACHE_DIR="/path/to/dir"` (see environment variables FAQ below)
   - Default OS cache directory:
     - Windows: `%APPDATA%\tabpfn\`
     - macOS: `~/Library/Caches/tabpfn/`
     - Linux: `~/.cache/tabpfn/`

**Q: I'm getting a `pickle` error when loading the model. What should I do?**
A: Try the following:
- Download the newest version of tabpfn `pip install tabpfn --upgrade`
- Ensure model files downloaded correctly (re-download if needed)

**Q: What environment variables can I use to configure TabPFN?**
A: TabPFN uses Pydantic settings for configuration, supporting environment variables and `.env` files:

**Model Configuration:**
- `TABPFN_MODEL_CACHE_DIR`: Custom directory for caching downloaded TabPFN models (default: platform-specific user cache directory)
- `TABPFN_ALLOW_CPU_LARGE_DATASET`: Allow running TabPFN on CPU with large datasets (>1000 samples). Set to `true` to override the CPU limitation. Note: This will be very slow!

**PyTorch Settings:**
- `PYTORCH_CUDA_ALLOC_CONF`: PyTorch CUDA memory allocation configuration to optimize GPU memory usage (default: `max_split_size_mb:512`). See [PyTorch CUDA documentation](https://docs.pytorch.org/docs/stable/notes/cuda.html#optimizing-memory-usage-with-pytorch-cuda-alloc-conf) for more information.

Example:
```bash
export TABPFN_MODEL_CACHE_DIR="/path/to/models"
export TABPFN_ALLOW_CPU_LARGE_DATASET=true
export PYTORCH_CUDA_ALLOC_CONF="max_split_size_mb:512"
```

Or simply set them in your `.env`

**Q: How do I save and load a trained TabPFN model?**
A: Use :func:`save_fitted_tabpfn_model` to persist a fitted estimator and reload
it later with :func:`load_fitted_tabpfn_model` (or the corresponding
``load_from_fit_state`` class methods).

```python
from tabpfn import TabPFNRegressor
from tabpfn.model_loading import (
    load_fitted_tabpfn_model,
    save_fitted_tabpfn_model,
)

# Train the regressor on GPU
reg = TabPFNRegressor(device="cuda")
reg.fit(X_train, y_train)
save_fitted_tabpfn_model(reg, "my_reg.tabpfn_fit")

# Later or on a CPU-only machine
reg_cpu = load_fitted_tabpfn_model("my_reg.tabpfn_fit", device="cpu")
```

To store just the foundation model weights (without a fitted estimator) use
``save_tabpfn_model(reg.model_, "my_tabpfn.ckpt")``. This merely saves a
checkpoint of the pre-trained weights so you can later create and fit a fresh
estimator. Reload the checkpoint with ``load_model_criterion_config``.

### **Performance & Limitations**

**Q: Can TabPFN handle missing values?**
A: **Yes!**

**Q: How can I improve TabPFN’s performance?**
A: Best practices:
- Use **AutoTabPFNClassifier** from [TabPFN Extensions](https://github.com/priorlabs/tabpfn-extensions) for post-hoc ensembling
- Feature engineering: Add domain-specific features to improve model performance

Not effective:
- Adapt feature scaling
- Convert categorical features to numerical values (e.g., one-hot encoding)

## Development

1. Setup environment:
```bash
python -m venv venv
source venv/bin/activate  # On Windows: venv\Scripts\activate
git clone https://github.com/PriorLabs/TabPFN.git
cd TabPFN
pip install -e ".[dev]"
pre-commit install
```

2. Before committing:
```bash
pre-commit run --all-files
```

3. Run tests:
```bash
pytest tests/
```

## Anonymized Telemetry

This project collects fully anonymous usage telemetry with an option to opt-out of any telemetry or opt-in to extended telemetry.

The data is used exclusively to help us provide stability to the relevant products and compute environments and guide future improvements.

- **No personal data is collected**
- **No code, model inputs, or outputs are ever sent**
- **Data is strictly anonymous and cannot be linked to individuals**

For details on telemetry, please see our [Telemetry Reference](https://github.com/PriorLabs/TabPFN/blob/main/TELEMETRY.md) and our [Privacy Policy](https://priorlabs.ai/privacy_policy/).

**To opt out**, set the following environment variable:

```bash
export TABPFN_DISABLE_TELEMETRY=1
```
---

Built with ❤️ by [Prior Labs](https://priorlabs.ai) - Copyright (c) 2025 Prior Labs GmbH

#' Controlling TabPFN execution
#'
#' @param n_preprocessing_jobs An integer for the number of worker processes.
#' A value of -1L indicates all possible resources.
#' @param device A character value for the device used for torch (e.g., `"cpu"`,
#' `"cuda"`, `"mps"`, etc.). Th default is `"auto"`.
#' @param ignore_pretraining_limits A logical to bypass the default data limits
#'   on:the number of training set samples (10,000) and, the number of
#'   predictors (500). There is an unchangeable limit to the number of classes
#'   (10).
#' @param inference_precision A character value for the trade off between speed
#' and reproducibility. This can be a torch `dtype`, `"autocast"` (for torch's
#' mixed-precision autocast), or "auto".
#' @param fit_mode A character value to control how the are preprocessed and/or
#' cached. Values are `"fit_preprocessors"` (the default), `"low_memory"`,
#' `"fit_with_cache"`, and `"batched"`.
#' @param memory_saving_mode A character string to help with out-of-memory
#' errors. Values are either a logical or `"auto"`.
#' @param random_state An integer to set the random number stream.
#' @param ... Additional named arguments passed directly to the TabPFN Python
#'   constructor. Use this to supply options not covered by the named parameters
#'   above (e.g. arguments added in newer versions of the Python package).
#' @return A list with extra class `"control_tab_pfn"` that has named elements
#' for each of the argument values.
#' @references
#' \url{https://github.com/PriorLabs/TabPFN/blob/main/src/tabpfn/classifier.py},
#' \url{https://github.com/PriorLabs/TabPFN/blob/main/src/tabpfn/regressor.py}
#' @examples
#' control_tab_pfn()
#' @export
control_tab_pfn <- function(
  n_preprocessing_jobs = 1L,
  device = "auto",
  ignore_pretraining_limits = FALSE,
  inference_precision = "auto",
  fit_mode = "fit_preprocessors",
  memory_saving_mode = "auto",
  random_state = sample.int(10^6, 1),
  ...
) {
  check_bool(ignore_pretraining_limits)
  check_string(fit_mode)
  check_string(device)
  check_string(inference_precision)
  check_number_whole(n_preprocessing_jobs)
  check_number_whole(random_state)

  fit_mode <- rlang::arg_match(
    fit_mode,
    c("fit_preprocessors", "low_memory", "fit_with_cache", "batched")
  )

  mem_msg <- "{.arg memory_saving_mode} should be a single logical or string."
  if (length(memory_saving_mode) != 1) {
    cli::cli_abort(mem_msg)
  }
  if (!is.character(memory_saving_mode) & !is.logical(memory_saving_mode)) {
    cli::cli_abort(mem_msg)
  }

  dot_args <- rlang::list2(...)
  reserved <- c(
    "n_preprocessing_jobs",
    "device",
    "ignore_pretraining_limits",
    "inference_precision",
    "fit_mode",
    "memory_saving_mode",
    "random_state"
  )
  conflicts <- intersect(names(dot_args), reserved)
  if (length(conflicts) > 0) {
    cli::cli_abort(
      "Argument{?s} {.arg {conflicts}} must be passed as named argument{?s}, not via {.code ...}."
    )
  }

  args <- c(
    list(
      n_preprocessing_jobs = as.integer(n_preprocessing_jobs),
      device = device,
      ignore_pretraining_limits = ignore_pretraining_limits,
      inference_precision = inference_precision,
      fit_mode = fit_mode,
      memory_saving_mode = memory_saving_mode,
      random_state = as.integer(random_state)
    ),
    dot_args
  )

  class(args) <- "control_tab_pfn"
  args
}

#' @export
print.control_tab_pfn <- function(x, ...) {
  defaults <- control_tab_pfn()
  common <- intersect(names(x), names(defaults))
  extra <- setdiff(names(x), names(defaults))
  non_default_common <- purrr::map2_lgl(
    x[common],
    defaults[common],
    ~ !identical(.x, .y)
  )
  non_default <- c(
    non_default_common,
    stats::setNames(rep(TRUE, length(extra)), extra)
  )

  cli::cli_inform("control object for {.fn tab_pfn}")
  if (any(non_default)) {
    cat("\n")
    cli::cli_inform("non-default arguments:")
    xsub <- x[non_default]
    lst <- purrr::map2(
      names(xsub),
      xsub,
      ~ cli::format_inline("{.arg {.x}}: {.val {.y}}")
    )
    names(lst) <- rep("*", length(lst))
    cli::cli_bullets(lst)
  }
  invisible(x)
}

test_that('control values', {
  set.seed(822)
  ctrl <- control_tab_pfn(
    n_preprocessing_jobs = 2L,
    device = "cpu",
    ignore_pretraining_limits = TRUE,
    inference_precision = "float32",
    fit_mode = "low_memory",
    memory_saving_mode = TRUE,
    random_state = 12345L
  )

  expect_s3_class(ctrl, "control_tab_pfn")
  expect_equal(ctrl$n_preprocessing_jobs, 2L)
  expect_equal(ctrl$device, "cpu")
  expect_equal(ctrl$ignore_pretraining_limits, TRUE)
  expect_equal(ctrl$inference_precision, "float32")
  expect_equal(ctrl$fit_mode, "low_memory")
  expect_equal(ctrl$memory_saving_mode, TRUE)
  expect_equal(ctrl$random_state, 12345L)

  set.seed(822)
  expect_snapshot(ctrl)

  set.seed(822)
  expect_snapshot(control_tab_pfn(random_state = 1))
})

test_that('control dot args - model_path', {
  skip_if_no_tabpfn()
  skip_if_not_installed("modeldata")

  tabpfn <- import_tabpfn()
  model_path <- tabpfn$TabPFNClassifier$create_default_for_version(
    "v2.6"
  )$model_path

  data(two_class_dat, package = "modeldata")
  x_tr <- two_class_dat[1:20, 1:2]
  y_tr <- two_class_dat$Class[1:20]

  set.seed(1)
  mod <- tab_pfn(x_tr, y_tr, control = control_tab_pfn(model_path = model_path))

  expect_s3_class(mod, exp_cls)
})

test_that('control dot args - invalid parameter fails', {
  skip_if_no_tabpfn()
  skip_if_not_installed("modeldata")

  data(two_class_dat, package = "modeldata")
  x_tr <- two_class_dat[1:20, 1:2]
  y_tr <- two_class_dat$Class[1:20]

  expect_error(
    tab_pfn(x_tr, y_tr, control = control_tab_pfn(not_a_real_param = TRUE))
  )
})

#' Predict using `TabPFN`
#'
#' @param object,x A `tab_pfn` object.
#'
#' @param new_data A data frame or matrix of new predictors.
#'
#' @param ... Not used, but required for extensibility.
#'
#' @return
#'
#' [predict()] returns a tibble of predictions and [augment()] appends the
#' columns in `new_data`. In either case, the number of rows in the tibble is
#' guaranteed to be the same as the number of rows in `new_data`.
#'
#' For regression data, the prediction is in the column `.pred`. For
#' classification, the class predictions are in `.pred_class` and the
#' probability estimates are in columns with the pattern `.pred_{level}` where
#' `level` is the levels of the outcome factor vector.
#'
#' @examples
#' # Minimal example for quick execution
#' car_train <- mtcars[ 1:5,   ]
#' car_test  <- mtcars[6, -1]
#'
#' \dontrun{
#' # Fit
#' if (is_tab_pfn_installed() & interactive()) {
#'  mod <- tab_pfn(mpg ~ cyl + log(drat), car_train)
#'
#'  # Predict
#'  predict(mod, car_test)
#'  augment(mod, car_test)
#' }
#' }
#'
#' @export
predict.tab_pfn <- function(object, new_data, ...) {
  rlang::check_dots_empty()
  forged <- hardhat::forge(new_data, object$blueprint)$predictors
  res <- predict(object$fit, forged, object$levels)
  res
}

# ------------------------------------------------------------------------------
# Implementation

#' @export
predict.tabpfn.regressor.TabPFNRegressor <- function(
  object,
  new_data,
  levels,
  ...
) {
  py_msg <- reticulate::py_capture_output(
    res <- try(object$predict(new_data), silent = TRUE)
  )

  if (inherits(res, "try-error")) {
    msgs <- as.character(res)
    cli::cli_abort("Prediction failed: {msgs}")
  } else {
    res <- tibble::tibble(.pred = as.vector(res))
  }

  res
}

#' @export
predict.tabpfn.classifier.TabPFNClassifier <- function(
  object,
  new_data,
  levels,
  ...
) {
  py_msg <- reticulate::py_capture_output(
    res <- try(object$predict_proba(new_data), silent = TRUE)
  )

  if (inherits(res, "try-error")) {
    msgs <- as.character(res)
    cli::cli_abort("Prediction failed: {msgs}")
  } else {
    colnames(res) <- paste0(".pred_", object$classes_)
    cls_ind <- apply(res, 1, which.max)
    res <- tibble::as_tibble(res)
    # TabPFN will reorder the class levels; if the original factor has levels "b"
    # and "a", object$classes_ will have c("a", "b)
    res$.pred_class <- factor(object$classes_[cls_ind], levels = levels)
  }

  res
}

#' @export
#' @rdname predict.tab_pfn
augment.tab_pfn <- function(x, new_data, ...) {
  new_data <- tibble::new_tibble(new_data)
  res <- predict(x, new_data)
  res <- cbind(res, new_data)
  tibble::new_tibble(res)
}

# Clean up any torchinductor directories after tests complete
torch_dirs <- list.files(
  path = getwd(),
  pattern = "^torchinductor",
  full.names = TRUE,
  recursive = FALSE,
  include.dirs = TRUE
)

for (dir in torch_dirs) {
  if (dir.exists(dir)) {
    unlink(dir, recursive = TRUE, force = TRUE)
  }
}

# Also clean up the test-specific temp cache
temp_cache <- file.path(tempdir(), "torchinductor_test")
if (dir.exists(temp_cache)) {
  unlink(temp_cache, recursive = TRUE, force = TRUE)
}

# nocov start
.pkg_env <- new.env()
.pkg_env$tab_pfn <- NULL

.onLoad <- function(...) {
  # Set PyTorch TorchInductor cache to R's temp directory
  # This prevents 'torchinductor' directory creation in working directory
  if (Sys.getenv("TORCHINDUCTOR_CACHE_DIR") == "") {
    torch_cache_dir <- file.path(tempdir(), "torchinductor")
    Sys.setenv(TORCHINDUCTOR_CACHE_DIR = torch_cache_dir)
  }

  reticulate::py_require("tabpfn")

  tryCatch(
    .pkg_env$tab_pfn <- reticulate::import(
      "tabpfn",
      delay_load = list(
        on_error = function(e) {
          cli::cli_abort(msg_tabpfn_not_available(e))
        },
        # See https://github.com/tidymodels/tabpfn/issues/3
        before_load = function() {
          check_libomp()
        }
      )
    ),

    # if reticulate has already loaded symbols from a Python installation,
    # `reticulate::import(delay_load = TRUE)` will error immediately.
    python.builtin.ModuleNotFoundError = function(e) {
      cli::cli_warn(msg_tabpfn_not_available(e))
    }
  )
}

.onUnload <- function(libpath) {
  # Clean up any torchinductor directories that may have been created
  # This is a defensive measure in case the environment variable didn't work

  # Check current working directory
  torch_dir_cwd <- file.path(getwd(), "torchinductor")
  if (dir.exists(torch_dir_cwd)) {
    unlink(torch_dir_cwd, recursive = TRUE, force = TRUE)
  }

  # Clean up in R CMD check temp directories (paths containing .Rcheck)
  if (grepl("\\.Rcheck", getwd(), fixed = TRUE)) {
    torch_dirs <- list.files(
      path = getwd(),
      pattern = "^torchinductor$",
      full.names = TRUE,
      recursive = FALSE,
      include.dirs = TRUE
    )
    for (dir in torch_dirs) {
      if (dir.exists(dir)) {
        unlink(dir, recursive = TRUE, force = TRUE)
      }
    }
  }
}

import_tabpfn <- function() {
  .pkg_env$tab_pfn
}

# nocov end

msg_tabpfn_not_available <- function(cnd) {
  c(
    x = "The {.pkg tabpfn} Python package is not installed in the discovered Python installation ({.file {reticulate::py_exe()}}).",
    i = 'Allow reticulate to automatically configure an ephemeral Python environment by
         removing the Python installation from the order of discovery and restarting the R session.
         See {.href [Order of Discovery](https://rstudio.github.io/reticulate/dev/articles/versions.html#order-of-discovery)} for more info.',
    # Or set env var {.code Sys.setenv("RETICULATE_USE_MANAGED_VENV" = "yes")}',
    i = 'Or install {.pkg tabpfn} into the selected Python environment by calling
        {.code reticulate::py_install("tabpfn")}'
  )
}

check_libomp <- function() {
  os_info <- Sys.info()[["sysname"]]
  if (os_info != "Darwin") {
    return(invisible(NULL))
  }

  vm_types <- system(paste("vmmap", Sys.getpid()), intern = TRUE)
  libomp_lines <- grep("libomp", vm_types, value = TRUE)

  if (length(libomp_lines) == 0) {
    return(invisible(NULL))
  }

  # Extract the file path from each vmmap line (path starts with "/" at end of line)
  libomp_paths <- sub(".*\\s(/\\S+)\\s*$", "\\1", libomp_lines)

  # libomp loaded from within the active Python environment is fine — torch will
  # reuse it. Only error if libomp came from outside the Python env (e.g. an R
  # package or an OpenMP-enabled R binary), because torch would then try to load
  # its own bundled copy alongside a foreign one, causing a segfault.
  py_env_root <- reticulate::py_config()$prefix
  outside_py_env <- !startsWith(libomp_paths, py_env_root)

  if (any(outside_py_env)) {
    cli::cli_abort(
      c(
        i = "We believe that an existing package has loaded {.pkg OpenMP}.",
        x = "{.pkg PyTorch} was about to do the same and would cause a segmentation fault.",
        i = "See {.url https://github.com/tidymodels/tabpfn/issues/3}.",
        "!" = "Try running {.code reticulate::import('torch')} in a new R session prior to loading other packages.",
        call = NULL
      )
    )
  }
  invisible(NULL)
}

# ------------------------------------------------------------------------------

row_limits <- 50000
col_limits <- 2000
cls_limits <- 10

check_data_constraints <- function(x, y, control) {
  lvls <- levels(y)

  x_dims <- dim(x)
  if (x_dims[1] > row_limits & !control$ignore_pretraining_limits) {
    cli::cli_abort(
      call = NULL,
      c(
        i = "There are {format(x_dims[1], big.mark = ',')} rows in the training set.",
        i = "TabPFN (version 2) is intended for training set sizes <=
        {format(row_limits, big.mark = ',')} rows.",
        i = "Consider setting the option {.arg ignore_pretraining_limits} to
        {.val TRUE} or subset the training size using the
        {.arg training_set_limit} argument."
      )
    )
  }
  if (x_dims[2] > col_limits & !control$ignore_pretraining_limits) {
    cli::cli_abort(
      call = NULL,
      c(
        i = "There are {format(x_dims[2], big.mark = ',')} predictors in the training set.",
        i = "TabPFN (version 2) is intended for <= {col_limits} predictors.",
        i = "Consider setting the option {.arg ignore_pretraining_limits} to {.val TRUE} or subset the training size."
      )
    )
  }

  if (!is.null(lvls) && length(lvls) > cls_limits) {
    cli::cli_abort(
      call = NULL,
      c(
        i = "There are {length(lvls)} classes in the outcome.",
        x = "TabPFN (version 2) is intended for <= {cls_limits} classes and won't work with more."
      )
    )
  }

  invisible(NULL)
}

# Sampling down the data for data constraints

sample_indicies <- function(molded, size_limit = row_limits) {
  num_rows <- nrow(molded$outcomes)
  if (num_rows <= size_limit) {
    return(integer(0))
  }

  dat <-
    molded$outcomes |>
    dplyr::mutate(.row_order = dplyr::row_number()) |>
    rlang::set_names(c("outcome", ".row_order"))

  is_factor <- is.factor(dat$outcome)

  if (is_factor) {
    data_subset <-
      dat |>
      dplyr::group_by(outcome) |>
      dplyr::group_nest(keep = TRUE) |>
      dplyr::mutate(
        size = purrr::map_int(data, nrow),
        sample_prop = size / num_rows,
        sample_num = ceiling(sample_prop * size_limit),
        data = purrr::map2(data, sample_num, ~ dplyr::slice_sample(.x, n = .y))
      )
  } else {
    data_subset <-
      dat |>
      dplyr::mutate(quantile = dplyr::ntile(outcome, n = 4)) |>
      dplyr::group_by(quantile) |>
      dplyr::group_nest(keep = TRUE) |>
      dplyr::mutate(
        size = purrr::map_int(data, nrow),
        sample_prop = size / num_rows,
        sample_num = ceiling(sample_prop * size_limit),
        data = purrr::map2(data, sample_num, ~ dplyr::slice_sample(.x, n = .y))
      )
  }

  purrr::map_dfr(data_subset$data, ~.x) |>
    dplyr::arrange(.row_order) |>
    dplyr::select(.row_order) |>
    dplyr::slice(1:size_limit) |>
    purrr::pluck(".row_order")
}

#' Check the Python package installation
#'
#' Attempts to import the Python package
#' @return A single logical
#' @examples
#' if (interactive()) {
#'  # This may take a minute
#'  is_tab_pfn_installed()
#' }
#' @export
is_tab_pfn_installed <- function() {
  suppressWarnings(
    res <- import_tabpfn() |>
      reticulate::py_has_attr("noexists") |> # Forcing load of package
      try(silent = TRUE)
  )
  !inherits(res, "try-error")
}


check_model_version <- function(x, call = rlang::caller_env()) {
  valid_versions <- tabpfn_list_versions()

  if (!x %in% valid_versions) {
    cli::cli_abort(
      c(
        "{.arg model_version} must be one of {.or {.val {valid_versions}}}.",
        x = "{.val {x}} is not a valid model version."
      ),
      call = call
    )
  }

  invisible(x)
}

test_that("check_model_version validates correctly", {
  skip_if_no_tabpfn()

  expect_no_error(check_model_version("v2"))
  expect_no_error(check_model_version("v2.5"))
  expect_error(check_model_version("V2"), "not a valid model version")
  expect_error(check_model_version("invalid"), "not a valid model version")
})

test_that("msg_tabpfn_not_available returns correct structure", {
  msg <- tabpfn:::msg_tabpfn_not_available()
  expect_named(msg, c("x", "i", "i"))
})

test_that("tab_pfn fits with all versions", {
  skip_if_no_tabpfn()
  for (version in tabpfn_list_versions()) {
    mod <- tab_pfn(am ~ mpg + wt, data = mtcars, version = version)
    expect_s3_class(mod, "tab_pfn")
  }
})

test_that("check_data_constraints errors when too many rows", {
  x <- matrix(0, nrow = 50001, ncol = 2)
  y <- factor(rep(c("a", "b"), length.out = 50001))
  expect_error(
    tabpfn:::check_data_constraints(x, y, control_tab_pfn()),
    "50,000"
  )
})

test_that("tab_pfn v2 errors with more than 500 predictors", {
  skip_if_no_tabpfn()
  n <- 100
  x <- as.data.frame(matrix(rnorm(n * 501), nrow = n, ncol = 501))
  x$y <- factor(rep(c("a", "b"), length.out = n))
  expect_error(
    tab_pfn(y ~ ., data = x, version = "v2")
  )
})

test_that("check_data_constraints errors when too many columns", {
  x <- matrix(0, nrow = 10, ncol = 2001)
  y <- factor(rep(c("a", "b"), length.out = 10))
  expect_error(
    tabpfn:::check_data_constraints(x, y, control_tab_pfn()),
    "2000"
  )
})

test_that("check_data_constraints errors when too many classes", {
  x <- matrix(0, nrow = 11, ncol = 2)
  y <- factor(letters[1:11])
  expect_error(
    tabpfn:::check_data_constraints(x, y, control_tab_pfn()),
    "classes"
  )
})

test_that("sample_indicies handles numeric outcomes", {
  set.seed(1)
  molded <- list(outcomes = data.frame(outcome = rnorm(50001)))
  result <- tabpfn:::sample_indicies(molded)
  expect_length(result, 50000)
  expect_true(all(result >= 1 & result <= 50001))
})

test_that("data constraints", {
  skip_if_no_tabpfn()
  skip_if_not_installed("modeldata")

  set.seed(418)
  orig_data <- tab_pfn(
    Class ~ .,
    data = modeldata::two_class_dat,
    num_estimators = 1,
  )

  expect_equal(orig_data$training[1], nrow(modeldata::two_class_dat))

  set.seed(418)
  smaller_data <- tab_pfn(
    Class ~ .,
    data = modeldata::two_class_dat,
    num_estimators = 1,
    training_set_limit = 50,
    control = control_tab_pfn(ignore_pretraining_limits = TRUE)
  )

  expect_equal(smaller_data$training[1], 50)
})

# This file is part of the standard setup for testthat.
# It is recommended that you do not modify it.
#
# Where should you do additional test configuration?
# Learn more about the roles of various files in:
# * https://r-pkgs.org/testing-design.html#sec-tests-files-overview
# * https://testthat.r-lib.org/articles/special-files.html

# For coverage testing set:
#   Sys.setenv("NOT_CRAN" = TRUE)

# To have Claude run coverage, get it to run:
# withr::with_envvar(
#  new = c("NOT_CRAN" = ""),
#  {
#   x <- covr::package_coverage(type = "tests")
#   file <- tempfile(fileext = ".html")
#   covr::report(x, file)
#   print(file)
#  }
# )

library(testthat)
library(tabpfn)

test_check("tabpfn")

new_tab_pfn <- function(
  fit,
  levels,
  training,
  logging,
  blueprint,
  call = NULL
) {
  cls <- c(
    "tabpfn.regressor.TabPFNRegressor",
    "tabpfn.classifier.TabPFNClassifier"
  )

  if (!inherits(fit, cls)) {
    cli::cli_abort(
      "The model fit object should have class {.cls {.or {cls}}}, not
			{.cls {class(fit)}}.",
      call = call
    )
  }

  check_character(levels, allow_null = TRUE)

  hardhat::new_model(
    fit = fit,
    levels = levels,
    training = training,
    logging = logging,
    blueprint = blueprint,
    class = "tab_pfn"
  )
}

#' Fit a TabPFN model.
#'
#' `tab_pfn()` applies data to a pre-estimated deep learning model defined by
#' Hollmann _et al_ (2025). This model emulates Bayesian inference for
#' regression and classification models.
#'
#' @param x Depending on the context:
#'
#'   * A __data frame__ of predictors.
#'   * A __matrix__ of predictors.
#'   * A __recipe__ specifying a set of preprocessing steps
#'     created from [recipes::recipe()].
#'
#' @param y When `x` is a __data frame__ or __matrix__, `y` is the outcome
#' specified as:
#'
#'   * A __data frame__ with 1 numeric column.
#'   * A __matrix__ with 1 numeric column.
#'   * A numeric __vector__ for regression or a __factor__ for classification.
#'
#' @param data When a __recipe__ or __formula__ is used, `data` is specified as:
#'
#'   * A __data frame__ containing both the predictors and the outcome.
#'
#' @param formula A formula specifying the outcome terms on the left-hand side,
#' and the predictor terms on the right-hand side.
#'
#' @param num_estimators An integer for the ensemble size. Default is `8L`.
#'
#' @param softmax_temperature An adjustment factor that is a divisor in the
#' exponents of the softmax function (see Details below). Defaults to 0.9.
#'
#' @param balance_probabilities A logical to adjust the prior probabilities in
#' cases where there is a class imbalance. Default is `FALSE`. Classification
#' only.
#'
#' @param average_before_softmax A logical. For cases where
#' `num_estimators > 1`, should the average be done before using the softmax
#' function or after? Default is `FALSE`.
#'
#' @param training_set_limit An integer greater than 2L (and possibly `Inf`)
#' that can be used to keep the training data within the limits of the
#' data constraints imposed by the Python library.
#'
#' @param version A character string for the model version (e.g., `"v2"`,
#' `"v2.5"`). When `NULL` (the default), the Python library's current default
#' version is used. When set, the model is initialized via
#' `create_default_for_version()` with the corresponding `ModelVersion` enum
#' value.
#'
#' @param control A list of options produced by [control_tab_pfn()].
#'
#' @param ... Not currently used, but required for extensibility.
#'
#' @details
#'
#' ## Computing Requirements
#'
#' This model can be used with or without a graphics processing unit (GPU).
#' However, it is fairly limited when used with a CPU (and no GPU). There might
#' be additional data size limitation warnings with CPU computations, and,
#' understandably, the execution time is much longer. CPU computations can also
#' consume a significant amount of system memory, depending on the size of your
#' data.
#'
#' GPUs using CUDA (Compute Unified Device Architecture) are most effective.
#' Limited testing with others has shown that GPUs with Metal Performance
#' Shaders (MPS) instructions (e.g., Apple GPUs) have limited utility for these
#' specific computations and might be slower than the CPU for some data sets.
#'
#' ## License Requirements
#'
#' Starting with version 2.5, using TabPFN requires accepting the model license
#' and obtaining a token from PriorLabs. Each model version (v2.5, v2.6, etc.)
#' has its own license that must be accepted individually.
#'
#' To set up access:
#'
#' 1. Visit [https://ux.priorlabs.ai](https://ux.priorlabs.ai) and create an
#'    account.
#' 2. Go to the **License** tab and accept the license for each model version
#'    you intend to use.
#' 3. Obtain your token from your account page.
#' 4. Set the `TABPFN_TOKEN` environment variable. The easiest way is to add it
#'    to your `.Renviron` file:
#'
#' \preformatted{
#' TABPFN_TOKEN=your_token_value
#' }
#'
#' The \pkg{usethis} function `edit_r_environ()` can be very helpful here.
#'
#' Users who already have `TABPFN_TOKEN` set can use TabPFN v2 without any
#' additional steps.
#'
#' ## Python Installation
#'
#' You will need a working Python virtual environment with the correct packages
#' to use these modeling functions.
#'
#' There are at least two ways to proceed.
#'
#' ### Ephemeral `uv` Install
#'
#' The first approach, which we *strongly suggest*, is to simply load this
#' package and attempt to run a model. This will prompt \pkg{reticulate} to
#' create an ephemeral environment and automatically install the required
#' packages. That process would look like this:
#'
#' \preformatted{
#'   > library(tabpfn)
#'   >
#'   > predictors <- mtcars[, -1]
#'   > outcome <- mtcars[, 1]
#'   >
#'   > # XY interface
#'   > mod <- tab_pfn(predictors, outcome)
#'   Downloading uv...Done!
#'   Downloading cpython-3.12.12 (download) (15.9MiB)
#'    Downloading cpython-3.12.12 (download)
#'   Downloading setuptools (1.1MiB)
#'   Downloading scikit-learn (8.2MiB)
#'   Downloading numpy (4.9MiB)
#'
#'   <downloading and installing more packages>
#'
#'    Downloading llvmlite
#'    Downloading torch
#'   Installed 58 packages in 350ms
#'   > mod
#'   TabPFN Regression Model
#'
#'   Training set
#'   i 32 data points
#'   i 10 predictors
#' }
#'
#' The location of the environment can be found at
#' `tools::R_user_dir("reticulate", "cache")`.
#'
#' See the documentation for [reticulate::py_require()] to learn more about this
#' method.
#'
#' ### Manually created `venv` Virtual Environment
#'
#' Alternatively, you can use the functions in the \pkg{reticulate} package to
#' create a virtual environment and install the required Python packages there.
#' An example pattern is:
#'
#' \preformatted{
#'   library(reticulate)
#'
#'   venv_name <- "r-tabpfn"    # exact name can be different
#'   venv_seed_python <-
#'     virtualenv_starter(">=3.11,<3.14") %||% install_python()
#'
#'   virtualenv_create(
#'     envname = venv_name,
#'     python = venv_seed_python,
#'     packages = c("numpy", "tabpfn")
#'   )
#' }
#'
#' Once you have that virtual environment installed, you can declare it as your
#' preferred Python installation with `use_virtualenv()`. (You must do this
#' before reticulate has initialized Python, i.e., before attempting to use
#' \pkg{tabpfn}):
#'
#' \preformatted{
#'   reticulate::use_virtualenv("r-tabpfn")
#' }
#'
#' ## Data
#'
#' Be default, there are limits to the training data dimensions:
#'
#'   * Version 2.0: number of training set samples (10,000) and, the number of
#'   predictors (500). There is an unchangeable limit to the number of classes
#'   (10).
#'
#'   * Version 2.5: number of training set samples (50,000) and, the number of
#'   predictors (2,000). There is an unchangeable limit to the number of classes
#'   (10).
#'
#' Predictors do not require preprocessing; missing values and factor vectors
#' are allowed.
#'
#' ## Model Selection
#'
#' By default, TabPFN uses the Python library's current default model version.
#' There are two ways to override this.
#'
#' ### Selecting a model version
#'
#' Use the `version` argument to select a specific released model version. For
#' example:
#'
#' \preformatted{
#'   # Use version 2.0
#'   mod <- tab_pfn(predictors, outcome, version = "v2")
#'
#'   # Use version 2.5
#'   mod <- tab_pfn(predictors, outcome, version = "v2.5")
#' }
#'
#' ### Pointing to a local model file
#'
#' If you have a model file on disk (e.g., downloaded for offline use), pass
#' its path via `control_tab_pfn(model_path = ...)`:
#'
#' \preformatted{
#'   ctrl <- control_tab_pfn(model_path = "/path/to/model_file.ckpt")
#'   mod  <- tab_pfn(predictors, outcome, control = ctrl)
#' }
#'
#' Note that `version` and `model_path` are mutually exclusive: if `version`
#' is set, it overwrites any `model_path` supplied through `control`.
#'
#' ## Calculations
#'
#' For the `softmax_temperature` value, the softmax terms are:
#'
#' \preformatted{
#' exp(value / softmax_temperature)
#' }
#'
#' A value of `softmax_temperature = 1` results in a plain softmax value.
#'
#' @return
#'
#' A `tab_pfn` object with elements:
#'
#'   * `fit`: the python object containing the model.
#'   * `levels`: a character string of class levels (or NULL for regression)
#'   * `training`: a vector with the training set dimensions.
#'   * `logging`: any R or python messages produced by the computations.
#'   * `blueprint`: am object produced by [hardhat::mold()] used to process
#'      new data during prediction.
#'
#' @references
#'
#' Hollmann, Noah, Samuel Müller, Lennart Purucker, Arjun Krishnakumar, Max
#' Körfer, Shi Bin Hoo, Robin Tibor Schirrmeister, and Frank Hutter.
#' "Accurate predictions on small data with a tabular foundation model."
#'  _Nature_ 637, no. 8045 (2025): 319-326.
#'
#' Hollmann, Noah, Samuel Müller, Katharina Eggensperger, and Frank Hutter.
#' "Tabpfn: A transformer that solves small tabular classification problems in
#' a second." _arXiv preprint_ arXiv:2207.01848 (2022).
#'
#' Müller, Samuel, Noah Hollmann, Sebastian Pineda Arango, Josif Grabocka, and
#' Frank Hutter. "Transformers can do Bayesian inference." _arXiv preprint_
#' arXiv:2112.10510 (2021).
#'
#' @seealso [control_tab_pfn()], [predict.tab_pfn()]
#' @examples
#' predictors <- mtcars[, -1]
#' outcome <- mtcars[, 1]
#'
#' \dontrun{
#' if (is_tab_pfn_installed() & interactive()) {
#'  # XY interface
#'  mod <- tab_pfn(predictors, outcome)
#'
#'  # Formula interface
#'  mod2 <- tab_pfn(mpg ~ ., mtcars)
#'
#'  # Recipes interface
#'  if (rlang::is_installed("recipes")) {
#'   suppressPackageStartupMessages(library(recipes))
#'   rec <-
#'    recipe(mpg ~ ., mtcars) %>%
#'    step_log(disp)
#'
#'   mod3 <- tab_pfn(rec, mtcars)
#'   mod3
#'  }
#' }
#' }
#'
#' @export
tab_pfn <- function(x, ...) {
  UseMethod("tab_pfn")
}

#' @export
#' @rdname tab_pfn
tab_pfn.default <- function(x, ...) {
  cli::cli_abort("{.fn tab_pfn} is not defined for {obj_type_friendly(x)}.")
}

# XY method - data frame

#' @export
#' @rdname tab_pfn
tab_pfn.data.frame <- function(
  x,
  y,
  num_estimators = 8L,
  softmax_temperature = 0.9,
  balance_probabilities = FALSE,
  average_before_softmax = FALSE,
  training_set_limit = 10000,
  version = NULL,
  control = control_tab_pfn(),
  ...
) {
  options <- control
  options$n_estimators <- num_estimators
  options$softmax_temperature <- softmax_temperature
  options$balance_probabilities <- balance_probabilities
  options$average_before_softmax <- average_before_softmax
  options <- check_fit_args(options)
  check_number_whole(training_set_limit, min = 2, allow_infinite = TRUE)

  processed <- hardhat::mold(x, y)
  tr_ind <- sample_indicies(processed, size_limit = training_set_limit)
  if (length(tr_ind) > 0) {
    processed$predictors <- processed$predictors[tr_ind, , drop = FALSE]
    processed$outcomes <- processed$outcomes[tr_ind, , drop = FALSE]
  }

  tab_pfn_bridge(processed, options, version = version, ...)
}

# XY method - matrix

#' @export
#' @rdname tab_pfn
tab_pfn.matrix <- function(
  x,
  y,
  num_estimators = 8L,
  softmax_temperature = 0.9,
  balance_probabilities = FALSE,
  average_before_softmax = FALSE,
  training_set_limit = 10000,
  version = NULL,
  control = control_tab_pfn(),
  ...
) {
  options <- control
  options$n_estimators <- num_estimators
  options$softmax_temperature <- softmax_temperature
  options$balance_probabilities <- balance_probabilities
  options$average_before_softmax <- average_before_softmax
  options <- check_fit_args(options)
  check_number_whole(training_set_limit, min = 2, allow_infinite = TRUE)

  processed <- hardhat::mold(x, y)
  tr_ind <- sample_indicies(processed, size_limit = training_set_limit)
  if (length(tr_ind) > 0) {
    processed$predictors <- processed$predictors[tr_ind, , drop = FALSE]
    processed$outcomes <- processed$outcomes[tr_ind, , drop = FALSE]
  }

  tab_pfn_bridge(processed, options, version = version, ...)
}

# Formula method

#' @export
#' @rdname tab_pfn
tab_pfn.formula <- function(
  formula,
  data,
  num_estimators = 8L,
  softmax_temperature = 0.9,
  balance_probabilities = FALSE,
  average_before_softmax = FALSE,
  training_set_limit = 10000,
  version = NULL,
  control = control_tab_pfn(),
  ...
) {
  options <- control
  options$n_estimators <- num_estimators
  options$softmax_temperature <- softmax_temperature
  options$balance_probabilities <- balance_probabilities
  options$average_before_softmax <- average_before_softmax
  options <- check_fit_args(options)
  check_number_whole(training_set_limit, min = 2, allow_infinite = TRUE)

  # No not convert factors to indicators:
  bp <- hardhat::default_formula_blueprint(
    intercept = FALSE,
    allow_novel_levels = FALSE,
    indicators = "none",
    composition = "tibble"
  )
  processed <- hardhat::mold(formula, data, blueprint = bp)
  tr_ind <- sample_indicies(processed, size_limit = training_set_limit)
  if (length(tr_ind) > 0) {
    processed$predictors <- processed$predictors[tr_ind, , drop = FALSE]
    processed$outcomes <- processed$outcomes[tr_ind, , drop = FALSE]
  }

  tab_pfn_bridge(processed, options, version = version, ...)
}

# Recipe method

#' @export
#' @rdname tab_pfn
tab_pfn.recipe <- function(
  x,
  data,
  num_estimators = 8L,
  softmax_temperature = 0.9,
  balance_probabilities = FALSE,
  average_before_softmax = FALSE,
  training_set_limit = 10000,
  version = NULL,
  control = control_tab_pfn(),
  ...
) {
  options <- control
  options$n_estimators <- num_estimators
  options$softmax_temperature <- softmax_temperature
  options$balance_probabilities <- balance_probabilities
  options$average_before_softmax <- average_before_softmax
  options <- check_fit_args(options)
  check_number_whole(training_set_limit, min = 2, allow_infinite = TRUE)

  processed <- hardhat::mold(x, data)
  tr_ind <- sample_indicies(processed, size_limit = training_set_limit)
  if (length(tr_ind) > 0) {
    processed$predictors <- processed$predictors[tr_ind, , drop = FALSE]
    processed$outcomes <- processed$outcomes[tr_ind, , drop = FALSE]
  }

  tab_pfn_bridge(processed, options, version = version, ...)
}

# ------------------------------------------------------------------------------
# Bridge

tab_pfn_bridge <- function(processed, options, version = NULL, ...) {
  rlang::check_dots_empty()

  if (!is.null(version)) {
    check_model_version(version)
  }

  predictors <- processed$predictors
  outcome <- processed$outcomes[[1]]

  check_data_constraints(predictors, outcome, options)

  res <- tab_pfn_impl(predictors, outcome, options, version = version)

  new_tab_pfn(
    fit = res$fit,
    levels = res$lvls,
    training = res$train,
    logging = res$logging,
    blueprint = processed$blueprint
  )
}

# ------------------------------------------------------------------------------
# Implementation

tab_pfn_impl <- function(x, y, opts, version = NULL) {
  tabpfn <- import_tabpfn()

  if (!is.null(version)) {
    if (is.factor(y)) {
      default_model <- tabpfn$TabPFNClassifier
    } else {
      default_model <- tabpfn$TabPFNRegressor
    }
    opts$model_path <- default_model$create_default_for_version(
      version
    )$model_path
  }

  cls_wrapper <- function(...) {
    tabpfn$TabPFNClassifier(...)
  }
  reg_wrapper <- function(...) {
    tabpfn$TabPFNRegressor(...)
  }

  if (is.factor(y)) {
    cls_cl <- rlang::call2("cls_wrapper", !!!opts)
    mod_obj <- rlang::eval_bare(cls_cl)
  } else if (is.numeric(y)) {
    opts <- opts[names(opts) != "balance_probabilities"]
    cls_cl <- rlang::call2("reg_wrapper", !!!opts)
    mod_obj <- rlang::eval_bare(cls_cl)
  }

  py_msg <- reticulate::py_capture_output(
    model_fit <- try(mod_obj$fit(x, y), silent = TRUE)
  )

  if (inherits(model_fit, "try-error")) {
    msgs <- as.character(model_fit)
    cli::cli_abort("Model failed: {msgs}")
  } else {
    msgs <- character(0)
  }

  # check for failures
  res <- list(
    fit = model_fit,
    lvls = levels(y),
    train = dim(x),
    logging = c(r = msgs, py = py_msg)
  )
  class(res) <- c("tab_pfn")
  res
}

#' @export
print.tab_pfn <- function(x, ...) {
  type <- ifelse(is.null(x$levels), "Regression", "Classification")
  cli::cli_inform("TabPFN {type} Model")
  cat("\n")
  cli::cli_inform("Training set\n\n")
  cli::cli_inform(c(i = "{x$training[1]} data point{?s}"))
  cli::cli_inform(c(i = "{x$training[2]} predictor{?s}"))

  if (!is.null(x$levels)) {
    cli::cli_inform(c(i = "class levels: {.val {x$levels}}"))
  }

  invisible(x)
}

check_fit_args <- function(opts, call = rlang::caller_env()) {
  check_number_whole(
    # These arg names are deliberately different
    opts$n_estimators,
    arg = "num_estimators",
    min = 1,
    call = call
  )
  opts$n_estimators <- as.integer(opts$n_estimators)

  check_number_decimal(
    opts$softmax_temperature,
    arg = "softmax_temperature",
    min = .Machine$double.eps,
    call = call
  )

  check_logical(
    opts$balance_probabilities,
    arg = "balance_probabilities",
    call = call
  )

  check_logical(
    opts$average_before_softmax,
    arg = "average_before_softmax",
    call = call
  )

  # ------------------------------------------------------------------------------
  # There have been some argument name differences in the python package versions

  arg_names <- names(opts)
  tabpfn <- import_tabpfn()
  py_arg_names <- names(formals(tabpfn$TabPFNClassifier))
  if (any(py_arg_names == "n_jobs")) {
    names(opts) <- gsub("^n_preprocessing_jobs$", "n_jobs", names(opts))
  }

  # ------------------------------------------------------------------------------

  opts
}

show_env_var <- function(x) {
  val <- Sys.getenv(x)
  if (identical(val, "")) {
    val <- "<not set>"
  }
  val
}

test_that('classification models', {
  skip_if_no_tabpfn()
  skip_if_not_installed("modeldata")

  #-----------------------------------------------------------------------------

  data(two_class_dat, package = "modeldata")
  x_tr_df <- two_class_dat[1:20, 1:2]
  x_tr_mat <- as.matrix(x_tr_df)
  y_tr <- two_class_dat$Class[1:20]

  x_te_df <- two_class_dat[21:23, 1:2]
  x_te_mat <- as.matrix(x_te_df)

  pred_ptype <-
    tibble::tibble(
      .pred_Class1 = numeric(0),
      .pred_Class2 = numeric(0),
      .pred_class = factor(0, levels = levels(two_class_dat$Class))
    )

  #-----------------------------------------------------------------------------

  set.seed(956)
  mod_df <- try(tab_pfn(x_tr_df, y_tr), silent = TRUE)
  expect_s3_class(mod_df, exp_cls)
  expect_snapshot(mod_df)

  pred_df <- predict(mod_df, x_te_df)
  expect_equal(pred_df[0, ], pred_ptype)
  expect_equal(nrow(pred_df), 3L)

  aug_df <- augment(mod_df, x_te_df)
  expect_s3_class(aug_df, c("tbl_df", "tbl", "data.frame"))
  expect_equal(nrow(aug_df), 3L)
  expect_equal(ncol(aug_df), 5L)

  #-----------------------------------------------------------------------------

  set.seed(956)
  mod_f <- try(tab_pfn(Class ~ ., data = two_class_dat[1:20, ]), silent = TRUE)
  expect_s3_class(mod_f, exp_cls)
  expect_snapshot(mod_f)

  pred_f <- predict(mod_f, x_tr_df[21:23, ])
  expect_equal(pred_f[0, ], pred_ptype)
  expect_equal(nrow(pred_f), 3L)

  aug_f <- augment(mod_f, x_tr_df[21:23, ])
  expect_s3_class(aug_f, c("tbl_df", "tbl", "data.frame"))
  expect_equal(nrow(aug_f), 3L)
  expect_equal(ncol(aug_f), 5L)

  #-----------------------------------------------------------------------------

  set.seed(956)
  mod_mat <- try(tab_pfn(x_tr_mat, y_tr), silent = TRUE)
  expect_s3_class(mod_mat, exp_cls)
  expect_snapshot(mod_mat)

  pred_mat <- predict(mod_mat, x_te_mat)
  expect_equal(pred_mat[0, ], pred_ptype)
  expect_equal(nrow(pred_mat), 3L)

  aug_mat <- augment(mod_mat, x_tr_df[21:23, ])
  expect_s3_class(aug_mat, c("tbl_df", "tbl", "data.frame"))
  expect_equal(nrow(aug_mat), 3L)
  expect_equal(ncol(aug_mat), 5L)
})

test_that('classification models - recipes', {
  skip_if_no_tabpfn()
  skip_if_not_installed("modeldata")
  skip_if_not_installed("recipes")

  reticulate::import("torch")

  library(tabpfn)
  suppressPackageStartupMessages(library(recipes))

  #-----------------------------------------------------------------------------

  data(two_class_dat, package = "modeldata")

  pred_ptype <-
    tibble::tibble(
      .pred_Class1 = numeric(0),
      .pred_Class2 = numeric(0),
      .pred_class = factor(0, levels = levels(two_class_dat$Class))
    )

  #-----------------------------------------------------------------------------

  rec <-
    recipe(Class ~ ., data = two_class_dat) |>
    step_interact(~ A:B)

  set.seed(956)
  mod_rec <- try(tab_pfn(rec, two_class_dat[1:20, ]), silent = TRUE)
  expect_s3_class(mod_rec, exp_cls)
  expect_snapshot(mod_rec)

  pred_rec <- predict(mod_rec, two_class_dat[50:52, ])
  expect_equal(pred_rec[0, ], pred_ptype)
  expect_equal(nrow(pred_rec), 3L)

  aug_rec <- augment(mod_rec, two_class_dat[50:52, ])
  expect_s3_class(aug_rec, c("tbl_df", "tbl", "data.frame"))
  expect_equal(nrow(aug_rec), 3L)
  expect_equal(ncol(aug_rec), 6L)
})

test_that('main options', {
  skip_if_no_tabpfn()
  set.seed(956)
  expect_snapshot_error(
    tab_pfn(Class ~ ., data = two_class_dat, num_estimators = "YES")
  )
  expect_snapshot_error(
    tab_pfn(Class ~ ., data = two_class_dat, softmax_temperature = -1)
  )
  expect_snapshot_error(
    tab_pfn(Class ~ ., data = two_class_dat, balance_probabilities = "nope")
  )
  expect_snapshot_error(
    tab_pfn(Class ~ ., data = two_class_dat, average_before_softmax = "suuuure")
  )
})

test_that("show_env_var returns value when set", {
  withr::with_envvar(
    new = c(TABPFN_TEST_VAR = "hello"),
    expect_equal(tabpfn:::show_env_var("TABPFN_TEST_VAR"), "hello")
  )
})

test_that("show_env_var returns not set when empty", {
  withr::with_envvar(
    new = c(TABPFN_TEST_VAR = ""),
    expect_equal(tabpfn:::show_env_var("TABPFN_TEST_VAR"), "<not set>")
  )
})

test_that('regression models', {
  skip_if_no_tabpfn()

  pred_ptype <- tibble::tibble(.pred = numeric(0))

  #-----------------------------------------------------------------------------

  set.seed(166)
  mod_df <- try(tab_pfn(predictors, outcome), silent = TRUE)
  expect_s3_class(mod_df, exp_cls)
  expect_snapshot(mod_df)

  pred_df <- predict(mod_df, mtcars[1:3, -1])
  expect_equal(pred_df[0, ], pred_ptype)
  expect_equal(nrow(pred_df), 3L)

  aug_df <- augment(mod_df, mtcars[1:3, -1])
  expect_s3_class(aug_df, c("tbl_df", "tbl", "data.frame"))
  expect_equal(nrow(aug_df), 3L)
  expect_equal(ncol(aug_df), 11L)

  #-----------------------------------------------------------------------------

  set.seed(166)
  mod_f <- try(tab_pfn(mpg ~ ., data = mtcars), silent = TRUE)
  expect_s3_class(mod_f, exp_cls)
  expect_snapshot(mod_f)

  pred_f <- predict(mod_f, mtcars[1:3, -1])
  expect_equal(pred_f[0, ], pred_ptype)
  expect_equal(nrow(pred_f), 3L)

  aug_f <- augment(mod_f, mtcars[1:3, -1])
  expect_s3_class(aug_f, c("tbl_df", "tbl", "data.frame"))
  expect_equal(nrow(aug_f), 3L)
  expect_equal(ncol(aug_f), 11L)

  #-----------------------------------------------------------------------------

  set.seed(166)
  mod_mat <- try(tab_pfn(as.matrix(predictors), outcome), silent = TRUE)
  expect_s3_class(mod_mat, exp_cls)
  expect_snapshot(mod_mat)

  pred_mat <- predict(mod_mat, mtcars[1:3, -1])
  expect_equal(pred_mat[0, ], pred_ptype)
  expect_equal(nrow(pred_mat), 3L)

  aug_mat <- augment(mod_mat, mtcars[1:3, -1])
  expect_s3_class(aug_mat, c("tbl_df", "tbl", "data.frame"))
  expect_equal(nrow(aug_mat), 3L)
  expect_equal(ncol(aug_mat), 11L)

  #-----------------------------------------------------------------------------

  expect_snapshot_error(
    tab_pfn(1, 2)
  )
})

test_that('training_set_limit with data frame and matrix interfaces', {
  skip_if_no_tabpfn()
  skip_if_not_installed("recipes")

  set.seed(166)
  mod_df <- try(
    tab_pfn(
      predictors,
      outcome,
      training_set_limit = 20,
      control = control_tab_pfn(ignore_pretraining_limits = TRUE)
    ),
    silent = TRUE
  )
  expect_equal(mod_df$training[1], 20)

  set.seed(166)
  mod_mat <- try(
    tab_pfn(
      as.matrix(predictors),
      outcome,
      training_set_limit = 20,
      control = control_tab_pfn(ignore_pretraining_limits = TRUE)
    ),
    silent = TRUE
  )
  expect_equal(mod_mat$training[1], 20)

  set.seed(166)
  rec <- recipes::recipe(mpg ~ ., data = mtcars)
  mod_rec <- try(
    tab_pfn(
      rec,
      mtcars,
      training_set_limit = 20,
      control = control_tab_pfn(ignore_pretraining_limits = TRUE)
    ),
    silent = TRUE
  )
  expect_equal(mod_rec$training[1], 20)
})

test_that('reproducible results', {
  skip_if_no_tabpfn()

  set.seed(166)
  mod_1 <- try(tab_pfn(predictors, outcome), silent = TRUE)
  pred_1 <- predict(mod_1, mtcars[1:3, -1])

  set.seed(166)
  mod_2 <- try(tab_pfn(predictors, outcome), silent = TRUE)
  pred_2 <- predict(mod_2, mtcars[1:3, -1])

  expect_equal(pred_1, pred_2)

  set.seed(774)
  mod_3 <- try(tab_pfn(predictors, outcome), silent = TRUE)
  pred_3 <- predict(mod_3, mtcars[1:3, -1])

  expect_false(all(pred_1$.pred == pred_3$.pred))

  set.seed(166)
  mod_4 <- try(tab_pfn(predictors, outcome, num_estimators = 1), silent = TRUE)
  pred_4 <- predict(mod_4, mtcars[1:3, -1])

  expect_false(all(pred_1$.pred == pred_4$.pred))
})


test_that('regression models - recipes', {
  skip_if_no_tabpfn()
  skip_if_not_installed("modeldata")
  skip_if_not_installed("recipes")

  reticulate::import("torch")

  library(tabpfn)
  library(recipes)
  data(Chicago, package = "modeldata")

  pred_ptype <- tibble::tibble(.pred = numeric(0))

  #-----------------------------------------------------------------------------

  rec <-
    recipe(ridership ~ Austin + Quincy_Wells + date, data = Chicago) |>
    step_date(date) |>
    step_rm(date)

  set.seed(166)
  mod_rec <- try(tab_pfn(rec, Chicago[1:20, ]), silent = TRUE)
  expect_s3_class(mod_rec, exp_cls)
  expect_snapshot(mod_rec)

  pred_rec <- predict(mod_rec, Chicago[50:52, ])
  expect_equal(pred_rec[0, ], pred_ptype)
  expect_equal(nrow(pred_rec), 3L)

  aug_rec <- augment(mod_rec, Chicago[50:52, ])
  expect_s3_class(aug_rec, c("tbl_df", "tbl", "data.frame"))
  expect_equal(nrow(aug_rec), 3L)
  expect_equal(ncol(aug_rec), 51L)
})

# Ensure PyTorch cache goes to temp directory during tests
# This is redundant with .onLoad() but provides extra safety for tests
if (Sys.getenv("TORCHINDUCTOR_CACHE_DIR") == "") {
  torch_cache_dir <- file.path(tempdir(), "torchinductor_test")
  Sys.setenv(TORCHINDUCTOR_CACHE_DIR = torch_cache_dir)
}

skip_if_no_tabpfn <- function() {
  skip_if(
    !is_tab_pfn_installed(),
    message = "TabPFN Python library is not installed"
  )
  skip_on_cran()
}

exp_cls <- c("tabpfn", "hardhat_model", "hardhat_scalar")

predictors <- mtcars[, -1]
outcome <- mtcars[, 1]

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import rlang
#' @importFrom stats predict

#' @importFrom generics augment
#' @export
generics::augment

## usethis namespace: end
NULL

utils::globalVariables(
  c(
    ".row_order",
    "data",
    "outcome",
    "quantile",
    "sample_num",
    "sample_prop",
    "size"
  )
)

#' List available TabPFN model versions
#'
#' Returns a character vector of valid model version strings accepted by
#' [tab_pfn()]'s `version` argument. The available model versions are queried
#' directly from the currently installed Python `tabpfn` library, not
#' hard-coded in this package, so results may differ across Python library
#' versions.
#' @return A character vector of model version strings.
#' @export
tabpfn_list_versions <- function() {
  if (!is_tab_pfn_installed()) {
    cli::cli_abort(msg_tabpfn_not_available())
  }

  tabpfn <- import_tabpfn()
  builtins <- reticulate::import_builtins()
  tabpfn$constants$ModelVersion |>
    builtins$list() |>
    unlist()
}

#' Download all TabPFN pre-trained model checkpoints
#'
#' @description
#' As of 2026-05-05, there are 36 pre-trained models equaling roughly 1.2 GB
#' of storage. Each model is trained on various synthetic & real datasets
#' tailored to classification & regression. This function routine will require
#' you to sign a one-time license for both 2.5 & 2.6 model varieties.
#' Downloading all models will take some time.
#'
#' @param cache_dir an option to override the default cache directory
#'
#' @returns Invisibly returns `NULL`. Called for its side effect of
#' downloading model files.
#'
#' @export
#' @examples
#' \donttest{
#' tabpfn_download_models()
#' }
tabpfn_download_models <- function(cache_dir = NULL) {
  pathlib <- reticulate::import("pathlib")
  tabpfn <- import_tabpfn()
  model_loading <- tabpfn$model_loading
  if (is.null(cache_dir)) {
    cache_dir <- model_loading$get_cache_dir()
  } else {
    cache_dir <- pathlib$Path(path.expand(cache_dir))
  }
  model_loading$download_all_models(to = cache_dir)
  invisible()
}

test_that("tabpfn_list_versions returns a character vector", {
  skip_if_no_tabpfn()

  versions <- tabpfn_list_versions()
  expect_type(versions, "character")
  expect_gt(length(versions), 0)
})

test_that("tabpfn_list_versions includes known versions", {
  skip_if_no_tabpfn()

  versions <- tabpfn_list_versions()
  expect_contains(versions, "v2")
})

test_that("tabpfn_list_versions errors when tabpfn not installed", {
  local_mocked_bindings(is_tab_pfn_installed = function() FALSE)
  expect_error(tabpfn_list_versions(), "not installed")
})

test_that("tabpfn_download_models runs without error", {
  skip_if_no_tabpfn()

  expect_no_error(tabpfn_download_models())
})

# control values

    Code
      ctrl
    Message
      control object for `tab_pfn()`
    Output
      
    Message
      non-default arguments:
      * `n_preprocessing_jobs`: 2
      * `device`: "cpu"
      * `ignore_pretraining_limits`: TRUE
      * `inference_precision`: "float32"
      * `fit_mode`: "low_memory"
      * `memory_saving_mode`: TRUE
      * `random_state`: 12345

---

    Code
      control_tab_pfn(random_state = 1)
    Message
      control object for `tab_pfn()`
    Output
      
    Message
      non-default arguments:
      * `random_state`: 1

# classification models

    Code
      mod_df
    Message
      TabPFN Classification Model
    Output
      
    Message
      Training set
      i 20 data points
      i 2 predictors
      i class levels: "Class1" and "Class2"

---

    Code
      mod_f
    Message
      TabPFN Classification Model
    Output
      
    Message
      Training set
      i 20 data points
      i 2 predictors
      i class levels: "Class1" and "Class2"

---

    Code
      mod_mat
    Message
      TabPFN Classification Model
    Output
      
    Message
      Training set
      i 20 data points
      i 2 predictors
      i class levels: "Class1" and "Class2"

# classification models - recipes

    Code
      mod_rec
    Message
      TabPFN Classification Model
    Output
      
    Message
      Training set
      i 20 data points
      i 3 predictors
      i class levels: "Class1" and "Class2"

# main options

    `num_estimators` must be a whole number, not the string "YES".

---

    `softmax_temperature` must be a number larger than or equal to 2.22044604925031e-16, not the number -1.

---

    `balance_probabilities` must be a logical vector, not the string "nope".

---

    `average_before_softmax` must be a logical vector, not the string "suuuure".

# regression models

    Code
      mod_df
    Message
      TabPFN Regression Model
    Output
      
    Message
      Training set
      i 32 data points
      i 10 predictors

---

    Code
      mod_f
    Message
      TabPFN Regression Model
    Output
      
    Message
      Training set
      i 32 data points
      i 10 predictors

---

    Code
      mod_mat
    Message
      TabPFN Regression Model
    Output
      
    Message
      Training set
      i 32 data points
      i 10 predictors

---

    `tab_pfn()` is not defined for the number 1.

# regression models - recipes

    Code
      mod_rec
    Message
      TabPFN Regression Model
    Output
      
    Message
      Training set
      i 20 data points
      i 5 predictors

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{is_tab_pfn_installed}
\alias{is_tab_pfn_installed}
\title{Check the Python package installation}
\usage{
is_tab_pfn_installed()
}
\value{
A single logical
}
\description{
Attempts to import the Python package
}
\examples{
if (interactive()) {
 # This may take a minute
 is_tab_pfn_installed()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabpfn-helpers.R
\name{tabpfn_list_versions}
\alias{tabpfn_list_versions}
\title{List available TabPFN model versions}
\usage{
tabpfn_list_versions()
}
\value{
A character vector of model version strings.
}
\description{
Returns a character vector of valid model version strings accepted by
\code{\link[=tab_pfn]{tab_pfn()}}'s \code{version} argument. The available model versions are queried
directly from the currently installed Python \code{tabpfn} library, not
hard-coded in this package, so results may differ across Python library
versions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{control_tab_pfn}
\alias{control_tab_pfn}
\title{Controlling TabPFN execution}
\usage{
control_tab_pfn(
  n_preprocessing_jobs = 1L,
  device = "auto",
  ignore_pretraining_limits = FALSE,
  inference_precision = "auto",
  fit_mode = "fit_preprocessors",
  memory_saving_mode = "auto",
  random_state = sample.int(10^6, 1),
  ...
)
}
\arguments{
\item{n_preprocessing_jobs}{An integer for the number of worker processes.
A value of -1L indicates all possible resources.}

\item{device}{A character value for the device used for torch (e.g., \code{"cpu"},
\code{"cuda"}, \code{"mps"}, etc.). Th default is \code{"auto"}.}

\item{ignore_pretraining_limits}{A logical to bypass the default data limits
on:the number of training set samples (10,000) and, the number of
predictors (500). There is an unchangeable limit to the number of classes
(10).}

\item{inference_precision}{A character value for the trade off between speed
and reproducibility. This can be a torch \code{dtype}, \code{"autocast"} (for torch's
mixed-precision autocast), or "auto".}

\item{fit_mode}{A character value to control how the are preprocessed and/or
cached. Values are \code{"fit_preprocessors"} (the default), \code{"low_memory"},
\code{"fit_with_cache"}, and \code{"batched"}.}

\item{memory_saving_mode}{A character string to help with out-of-memory
errors. Values are either a logical or \code{"auto"}.}

\item{random_state}{An integer to set the random number stream.}

\item{...}{Additional named arguments passed directly to the TabPFN Python
constructor. Use this to supply options not covered by the named parameters
above (e.g. arguments added in newer versions of the Python package).}
}
\value{
A list with extra class \code{"control_tab_pfn"} that has named elements
for each of the argument values.
}
\description{
Controlling TabPFN execution
}
\examples{
control_tab_pfn()
}
\references{
\url{https://github.com/PriorLabs/TabPFN/blob/main/src/tabpfn/classifier.py},
\url{https://github.com/PriorLabs/TabPFN/blob/main/src/tabpfn/regressor.py}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabpfn-helpers.R
\name{tabpfn_download_models}
\alias{tabpfn_download_models}
\title{Download all TabPFN pre-trained model checkpoints}
\usage{
tabpfn_download_models(cache_dir = NULL)
}
\arguments{
\item{cache_dir}{an option to override the default cache directory}
}
\value{
Invisibly returns \code{NULL}. Called for its side effect of
downloading model files.
}
\description{
As of 2026-05-05, there are 36 pre-trained models equaling roughly 1.2 GB
of storage. Each model is trained on various synthetic & real datasets
tailored to classification & regression. This function routine will require
you to sign a one-time license for both 2.5 & 2.6 model varieties.
Downloading all models will take some time.
}
\examples{
\donttest{
tabpfn_download_models()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabpfn-package.R
\docType{package}
\name{tabpfn-package}
\alias{tabpfn}
\alias{tabpfn-package}
\title{tabpfn: Prior-Data Fitted Network Foundational Model for Tabular Data}
\description{
Provides a consistent API for classification and regression models based on the 'TabPFN' model of Hollmann et al. (2025), "Accurate predictions on small data with a tabular foundation model," Nature, 637(8045) \doi{10.1038/s41586-024-08328-6}. The calculations are served via 'Python' to train and predict the model.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://tabpfn.tidymodels.org}
  \item \url{https://github.com/tidymodels/tabpfn}
  \item Report bugs at \url{https://github.com/tidymodels/tabpfn/issues}
}

}
\author{
\strong{Maintainer}: Max Kuhn \email{max@posit.co} (\href{https://orcid.org/0000-0003-2402-136X}{ORCID})

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabpfn-predict.R
\name{predict.tab_pfn}
\alias{predict.tab_pfn}
\alias{augment.tab_pfn}
\title{Predict using \code{TabPFN}}
\usage{
\method{predict}{tab_pfn}(object, new_data, ...)

\method{augment}{tab_pfn}(x, new_data, ...)
}
\arguments{
\item{object, x}{A \code{tab_pfn} object.}

\item{new_data}{A data frame or matrix of new predictors.}

\item{...}{Not used, but required for extensibility.}
}
\value{
\code{\link[=predict]{predict()}} returns a tibble of predictions and \code{\link[=augment]{augment()}} appends the
columns in \code{new_data}. In either case, the number of rows in the tibble is
guaranteed to be the same as the number of rows in \code{new_data}.

For regression data, the prediction is in the column \code{.pred}. For
classification, the class predictions are in \code{.pred_class} and the
probability estimates are in columns with the pattern \verb{.pred_\{level\}} where
\code{level} is the levels of the outcome factor vector.
}
\description{
Predict using \code{TabPFN}
}
\examples{
# Minimal example for quick execution
car_train <- mtcars[ 1:5,   ]
car_test  <- mtcars[6, -1]

\dontrun{
# Fit
if (is_tab_pfn_installed() & interactive()) {
 mod <- tab_pfn(mpg ~ cyl + log(drat), car_train)

 # Predict
 predict(mod, car_test)
 augment(mod, car_test)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TabPFN-fit.R
\name{tab_pfn}
\alias{tab_pfn}
\alias{tab_pfn.default}
\alias{tab_pfn.data.frame}
\alias{tab_pfn.matrix}
\alias{tab_pfn.formula}
\alias{tab_pfn.recipe}
\title{Fit a TabPFN model.}
\usage{
tab_pfn(x, ...)

\method{tab_pfn}{default}(x, ...)

\method{tab_pfn}{data.frame}(
  x,
  y,
  num_estimators = 8L,
  softmax_temperature = 0.9,
  balance_probabilities = FALSE,
  average_before_softmax = FALSE,
  training_set_limit = 10000,
  version = NULL,
  control = control_tab_pfn(),
  ...
)

\method{tab_pfn}{matrix}(
  x,
  y,
  num_estimators = 8L,
  softmax_temperature = 0.9,
  balance_probabilities = FALSE,
  average_before_softmax = FALSE,
  training_set_limit = 10000,
  version = NULL,
  control = control_tab_pfn(),
  ...
)

\method{tab_pfn}{formula}(
  formula,
  data,
  num_estimators = 8L,
  softmax_temperature = 0.9,
  balance_probabilities = FALSE,
  average_before_softmax = FALSE,
  training_set_limit = 10000,
  version = NULL,
  control = control_tab_pfn(),
  ...
)

\method{tab_pfn}{recipe}(
  x,
  data,
  num_estimators = 8L,
  softmax_temperature = 0.9,
  balance_probabilities = FALSE,
  average_before_softmax = FALSE,
  training_set_limit = 10000,
  version = NULL,
  control = control_tab_pfn(),
  ...
)
}
\arguments{
\item{x}{Depending on the context:
\itemize{
\item A \strong{data frame} of predictors.
\item A \strong{matrix} of predictors.
\item A \strong{recipe} specifying a set of preprocessing steps
created from \code{\link[recipes:recipe]{recipes::recipe()}}.
}}

\item{...}{Not currently used, but required for extensibility.}

\item{y}{When \code{x} is a \strong{data frame} or \strong{matrix}, \code{y} is the outcome
specified as:
\itemize{
\item A \strong{data frame} with 1 numeric column.
\item A \strong{matrix} with 1 numeric column.
\item A numeric \strong{vector} for regression or a \strong{factor} for classification.
}}

\item{num_estimators}{An integer for the ensemble size. Default is \code{8L}.}

\item{softmax_temperature}{An adjustment factor that is a divisor in the
exponents of the softmax function (see Details below). Defaults to 0.9.}

\item{balance_probabilities}{A logical to adjust the prior probabilities in
cases where there is a class imbalance. Default is \code{FALSE}. Classification
only.}

\item{average_before_softmax}{A logical. For cases where
\code{num_estimators > 1}, should the average be done before using the softmax
function or after? Default is \code{FALSE}.}

\item{training_set_limit}{An integer greater than 2L (and possibly \code{Inf})
that can be used to keep the training data within the limits of the
data constraints imposed by the Python library.}

\item{version}{A character string for the model version (e.g., \code{"v2"},
\code{"v2.5"}). When \code{NULL} (the default), the Python library's current default
version is used. When set, the model is initialized via
\code{create_default_for_version()} with the corresponding \code{ModelVersion} enum
value.}

\item{control}{A list of options produced by \code{\link[=control_tab_pfn]{control_tab_pfn()}}.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a \strong{recipe} or \strong{formula} is used, \code{data} is specified as:
\itemize{
\item A \strong{data frame} containing both the predictors and the outcome.
}}
}
\value{
A \code{tab_pfn} object with elements:
\itemize{
\item \code{fit}: the python object containing the model.
\item \code{levels}: a character string of class levels (or NULL for regression)
\item \code{training}: a vector with the training set dimensions.
\item \code{logging}: any R or python messages produced by the computations.
\item \code{blueprint}: am object produced by \code{\link[hardhat:mold]{hardhat::mold()}} used to process
new data during prediction.
}
}
\description{
\code{tab_pfn()} applies data to a pre-estimated deep learning model defined by
Hollmann \emph{et al} (2025). This model emulates Bayesian inference for
regression and classification models.
}
\details{
\subsection{Computing Requirements}{

This model can be used with or without a graphics processing unit (GPU).
However, it is fairly limited when used with a CPU (and no GPU). There might
be additional data size limitation warnings with CPU computations, and,
understandably, the execution time is much longer. CPU computations can also
consume a significant amount of system memory, depending on the size of your
data.

GPUs using CUDA (Compute Unified Device Architecture) are most effective.
Limited testing with others has shown that GPUs with Metal Performance
Shaders (MPS) instructions (e.g., Apple GPUs) have limited utility for these
specific computations and might be slower than the CPU for some data sets.
}

\subsection{License Requirements}{

Starting with version 2.5, using TabPFN requires accepting the model license
and obtaining a token from PriorLabs. Each model version (v2.5, v2.6, etc.)
has its own license that must be accepted individually.

To set up access:
\enumerate{
\item Visit \url{https://ux.priorlabs.ai} and create an
account.
\item Go to the \strong{License} tab and accept the license for each model version
you intend to use.
\item Obtain your token from your account page.
\item Set the \code{TABPFN_TOKEN} environment variable. The easiest way is to add it
to your \code{.Renviron} file:
}

\preformatted{
TABPFN_TOKEN=your_token_value
}

The \pkg{usethis} function \code{edit_r_environ()} can be very helpful here.

Users who already have \code{TABPFN_TOKEN} set can use TabPFN v2 without any
additional steps.
}

\subsection{Python Installation}{

You will need a working Python virtual environment with the correct packages
to use these modeling functions.

There are at least two ways to proceed.
\subsection{Ephemeral \code{uv} Install}{

The first approach, which we \emph{strongly suggest}, is to simply load this
package and attempt to run a model. This will prompt \pkg{reticulate} to
create an ephemeral environment and automatically install the required
packages. That process would look like this:

\preformatted{
  > library(tabpfn)
  >
  > predictors <- mtcars[, -1]
  > outcome <- mtcars[, 1]
  >
  > # XY interface
  > mod <- tab_pfn(predictors, outcome)
  Downloading uv...Done!
  Downloading cpython-3.12.12 (download) (15.9MiB)
   Downloading cpython-3.12.12 (download)
  Downloading setuptools (1.1MiB)
  Downloading scikit-learn (8.2MiB)
  Downloading numpy (4.9MiB)

  <downloading and installing more packages>

   Downloading llvmlite
   Downloading torch
  Installed 58 packages in 350ms
  > mod
  TabPFN Regression Model

  Training set
  i 32 data points
  i 10 predictors
}

The location of the environment can be found at
\code{tools::R_user_dir("reticulate", "cache")}.

See the documentation for \code{\link[reticulate:py_require]{reticulate::py_require()}} to learn more about this
method.
}

\subsection{Manually created \code{venv} Virtual Environment}{

Alternatively, you can use the functions in the \pkg{reticulate} package to
create a virtual environment and install the required Python packages there.
An example pattern is:

\preformatted{
  library(reticulate)

  venv_name <- "r-tabpfn"    # exact name can be different
  venv_seed_python <-
    virtualenv_starter(">=3.11,<3.14") %||% install_python()

  virtualenv_create(
    envname = venv_name,
    python = venv_seed_python,
    packages = c("numpy", "tabpfn")
  )
}

Once you have that virtual environment installed, you can declare it as your
preferred Python installation with \code{use_virtualenv()}. (You must do this
before reticulate has initialized Python, i.e., before attempting to use
\pkg{tabpfn}):

\preformatted{
  reticulate::use_virtualenv("r-tabpfn")
}
}

}

\subsection{Data}{

Be default, there are limits to the training data dimensions:
\itemize{
\item Version 2.0: number of training set samples (10,000) and, the number of
predictors (500). There is an unchangeable limit to the number of classes
(10).
\item Version 2.5: number of training set samples (50,000) and, the number of
predictors (2,000). There is an unchangeable limit to the number of classes
(10).
}

Predictors do not require preprocessing; missing values and factor vectors
are allowed.
}

\subsection{Model Selection}{

By default, TabPFN uses the Python library's current default model version.
There are two ways to override this.
\subsection{Selecting a model version}{

Use the \code{version} argument to select a specific released model version. For
example:

\preformatted{
  # Use version 2.0
  mod <- tab_pfn(predictors, outcome, version = "v2")

  # Use version 2.5
  mod <- tab_pfn(predictors, outcome, version = "v2.5")
}
}

\subsection{Pointing to a local model file}{

If you have a model file on disk (e.g., downloaded for offline use), pass
its path via \code{control_tab_pfn(model_path = ...)}:

\preformatted{
  ctrl <- control_tab_pfn(model_path = "/path/to/model_file.ckpt")
  mod  <- tab_pfn(predictors, outcome, control = ctrl)
}

Note that \code{version} and \code{model_path} are mutually exclusive: if \code{version}
is set, it overwrites any \code{model_path} supplied through \code{control}.
}

}

\subsection{Calculations}{

For the \code{softmax_temperature} value, the softmax terms are:

\preformatted{
exp(value / softmax_temperature)
}

A value of \code{softmax_temperature = 1} results in a plain softmax value.
}
}
\examples{
predictors <- mtcars[, -1]
outcome <- mtcars[, 1]

\dontrun{
if (is_tab_pfn_installed() & interactive()) {
 # XY interface
 mod <- tab_pfn(predictors, outcome)

 # Formula interface
 mod2 <- tab_pfn(mpg ~ ., mtcars)

 # Recipes interface
 if (rlang::is_installed("recipes")) {
  suppressPackageStartupMessages(library(recipes))
  rec <-
   recipe(mpg ~ ., mtcars) \%>\%
   step_log(disp)

  mod3 <- tab_pfn(rec, mtcars)
  mod3
 }
}
}

}
\references{
Hollmann, Noah, Samuel Müller, Lennart Purucker, Arjun Krishnakumar, Max
Körfer, Shi Bin Hoo, Robin Tibor Schirrmeister, and Frank Hutter.
"Accurate predictions on small data with a tabular foundation model."
\emph{Nature} 637, no. 8045 (2025): 319-326.

Hollmann, Noah, Samuel Müller, Katharina Eggensperger, and Frank Hutter.
"Tabpfn: A transformer that solves small tabular classification problems in
a second." \emph{arXiv preprint} arXiv:2207.01848 (2022).

Müller, Samuel, Noah Hollmann, Sebastian Pineda Arango, Josif Grabocka, and
Frank Hutter. "Transformers can do Bayesian inference." \emph{arXiv preprint}
arXiv:2112.10510 (2021).
}
\seealso{
\code{\link[=control_tab_pfn]{control_tab_pfn()}}, \code{\link[=predict.tab_pfn]{predict.tab_pfn()}}
}
